# DF2Tex is used to convert the given dataframe into LatexFile.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.extras.graph import DF2Tex
#
#     obj = DF2Tex()
#
#     obj.generateLatexCode(result, "minSup", "runtime", "algorithmColumn")
#
#     obj.print()
#
#     obj.save("outputFile.tex")
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import pandas as pd


class DF2Tex:
    """
    :Description:  generateLatexFileFromDatFrame is used to convert the given dataframe into LatexFile.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.graph import DF2Tex

            obj = DF2Tex()

            obj.generateLatexCode(result, "minSup", "runtime", "algorithmColumn")

            obj.print()

            obj.save("outputFile.tex")

    """

    def generateLatexCode(self, result: pd.DataFrame, xColumn, yColumn, algorithmColumn=None) -> None:
        titles = [xColumn, yColumn]
        legendary = pd.unique(result.iloc[:, 0].values.ravel())
        xaxisValues = result[xColumn].values
        yaxisValues = result[yColumn].values
        if algorithmColumn is None:
            algo = result.iloc[:, 0].values
        else:
            algo = result[algorithmColumn].values
        xLabel = titles[0]
        color = ['red', 'blue', 'green', 'black', 'yellow']
        latexCode = ""
        latexCode += "\\begin{axis}[\n\txlabel={\\Huge{" + xLabel + "}},"
        latexCode += "\n\tylabel={\\Huge{" + titles[1] + "}},"
        latexCode += "\n\txmin=" + str(min(xaxisValues)) + ", xmax=" + str(max(xaxisValues)) + ",]\n"
        for num in range(0, len(legendary)):
            latexCode += "\\addplot+  [" + color[num] + "]\n\tcoordinates {\n"
            for num2 in range(0, len(xaxisValues)):
                if legendary[num] == algo[num2]:
                    latexCode += "(" + str(xaxisValues[num2]) + "," + str(yaxisValues[num2]) + ")\n"
            latexCode += "\t};   \\addlegendentry{" + legendary[num] + "}\n"
            if num + 1 == len(legendary):
                latexCode += "\\end{axis}"
        DF2Tex.latexCode = latexCode

    def print(self):
        print(DF2Tex.latexCode)

    def save(outputFileName):
        with open(outputFileName, "w") as latexwriter:
            latexwriter.write(DF2Tex.latexCode)
        print(f"Latex file saved as {outputFileName}")


# Example usage
result = pd.DataFrame()
obj = DF2Tex()
# generateLatexCode function as four parameters dataFrame, xColumn-name, yColumn-name,
# algorithmColumn-name is optional
obj.generateLatexCode(result, "minSup", "runtime", "algorithmColumn")
# printLatexCode function prints the output of latex file
obj.print()
# save function gives the outputFile
obj.save("outputFile.tex")




# SequentialDatabase is to get stats of database like avarage, minimun, maximum  and so on
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.stats import sequentialDatabase as db
#
#             obj = db.sequentialDatabase(iFile, "\t")
#
#             obj.save(oFile)
#
#             obj.run()
#
#             obj.printStats()
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import statistics
import validators
from urllib.request import urlopen
import PAMI.extras.graph.plotLineGraphFromDictionary as plt
import sys
from typing import List, Dict, Tuple, Set, Union, Any, Generator


class sequentialDatabase():
    """
    SequentialDatabase is to get stats of database like avarage, minimun, maximum  and so on.

    :Attributes:

        :param inputFile: file :
               input file path
        :param sep: str
               separator in file. Default is tab space.

    :Methods:

        readDatabase():
            read sequential database from input file and store into database and size of each sequence and subsequences.
        getDatabaseSize(self):
            get the size of database
        getTotalNumberOfItems(self):
            get the number of items in database.
        getMinimumSequenceLength(self):
            get the minimum sequence length
        getAverageSubsequencePerSequenceLength(self):
            get the average subsequence length per sequence length. It is sum of all subsequence length divided by sequence length.
        getAverageItemPerSubsequenceLength(self):
            get the average Item length per subsequence. It is sum of all item length divided by subsequence length.
        getMaximumSequenceLength(self):
            get the maximum sequence length
        getStandardDeviationSubsequenceLength(self):
            get the standard deviation subsequence length
        getVarianceSequenceLength(self):
            get the variance Sequence length
        getSequenceSize(self):
            get the size of sequence
        getMinimumSubsequenceLength(self):
            get the minimum subsequence length
        getAverageItemPerSequenceLength(self):
            get the average item length per sequence. It is sum of all item length divided by sequence length.
        getMaximumSubsequenceLength(self):
            get the maximum subsequence length
        getStandardDeviationSubsequenceLength(self):
            get the standard deviation subsequence length
        getVarianceSubsequenceLength(self):
            get the variance subSequence length
        getSortedListOfItemFrequencies(self):
            get sorted list of item frequencies
        getFrequenciesInRange(self):
            get sorted list of item frequencies in some range
        getSequencialLengthDistribution(self):
            get Sequence length Distribution
        getSubsequencialLengthDistribution(self):
            get subSequence length distribution
        printStats(self):
            to print the all status of sequence database
        plotGraphs(self):
            to plot the  distribution about items, subsequences in sequence and items in subsequence

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.dbStats import SequentialDatabase as db

            obj = db.SequentialDatabase(iFile, "\t")

            obj.save(oFile)

            obj.run()

            obj.printStats()


    **Executing the code on terminal:**
    -------------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 sequentialDatabase.py <inputFile>

      Example Usage:

      (.venv) $ python3 sequentialDatabase.py sampleDB.txt



    **Sample run of the importing code:**
    ----------------------------------------------------
        import PAMI.extra.DBstats.SequentialDatabase as alg
        _ap=alg.SequentialDatabase(inputfile,sep)
        _ap.readDatabase()
        _ap.printStats()
        _ap.plotGraphs()
    **Credits:**
    ---------------------
        The complete program was written by Shota Suzuki  under the supervision of Professor Rage Uday Kiran.
    """

    def __init__(self, inputFile: str, sep: str='\t') -> None:
        """
        :param inputFile: input file name or path
        :type inputFile: str
        :param sep: separator character for input file
        :type sep: str
        :return: None
        """
        self.inputFile = inputFile
        self.seqLengthList = []
        self.subSeqLengthList = []
        self.sep = sep
        self.database = {}

    def readDatabase(self) -> None:
        """
        read sequential database from input file and store into database and size of each sequence and subsequences.
        """
        if isinstance(self.inputFile, str):
            if validators.url(self.inputFile):
                data = urlopen(self.inputFile)
                rowNum=0
                for line in data:
                    line.strip()
                    temp = [i.rstrip() for i in line.split('-1')]
                    temp = [x for x in temp if x]
                    temp.pop()
                    seq = []
                    self.seqLengthList.append(len(temp))
                    self.subSeqLengthList.append([len(i) for i in temp])
                    for i in temp:
                        if len(i) > 1:
                            tempSorted=list(sorted(set(i.split())))
                            seq.append(tempSorted)
                        else:
                            seq.append(i)
                    rowNum+=1
                    if seq:
                        self.database[rowNum]=seq
            else:
                with open(self.inputFile, 'r') as f:
                    rowNum = 0
                    for line in f:
                        temp = [i.rstrip(self.sep) for i in line.split('-1')]
                        temp = [x for x in temp if x]
                        temp.pop()
                        seq = []
                        self.seqLengthList.append(len(temp))
                        subseq=[]
                        for i in temp:
                            if len(i) > 1:
                                tempSorted = list(sorted(set(i.split())))
                                subseq.append(len(tempSorted))
                                seq.append(tempSorted)
                            else:
                                seq.append(i)
                                subseq.append(len(i))
                        if subseq!=[]:
                            self.subSeqLengthList.append(subseq)
                        rowNum += 1
                        if seq:
                            self.database[rowNum] = seq


    def getDatabaseSize(self) -> int:
        """
        get the size of database
        :return: dataset size
        :rtype: int
        """
        return len(self.database)

    def getTotalNumberOfItems(self) -> int:
        """
        get the number of items in database.
        :return: number of items
        :rtype: int
        """
        return len(self.getSortedListOfItemFrequencies())

    def getMinimumSequenceLength(self) -> int:
        """
        get the minimum sequence length
        :return: minimum sequence length
        :rtype: int
        """
        return min(self.seqLengthList)

    def getAverageSubsequencePerSequenceLength(self) -> float:
        """
        get the average subsequence length per sequence length. It is sum of all subsequence length divided by sequence length.
        :return: average subsequence length per sequence length
        :rtype: float
        """
        totalLength = sum(self.seqLengthList)
        return totalLength / len(self.database)

    def getAverageItemPerSubsequenceLength(self) -> float:

        """
        get the average Item length per subsequence. It is sum of all item length divided by subsequence length.
        :return: average Item length per subsequence
        :rtype: float
        """

        totalLength = sum(list(map(sum,self.subSeqLengthList)))
        return totalLength / sum(self.seqLengthList)

    def getMaximumSequenceLength(self) -> int:
        """
        get the maximum sequence length
        :return: maximum sequence length
        :rtype: int
        """
        return max(self.seqLengthList)

    def getStandardDeviationSequenceLength(self) -> float:
        """
        get the standard deviation sequence length
        :return: standard deviation sequence length
        :rtype: float
        """
        return statistics.pstdev(self.seqLengthList)

    def getVarianceSequenceLength(self) -> float:
        """
        get the variance Sequence length
        :return: variance Sequence length
        :rtype: float
        """
        return statistics.variance(self.seqLengthList)

    def getSequenceSize(self) -> int:
        """
        get the size of sequence
        :return: sequences size
        :rtype: int
        """
        return sum(self.seqLengthList)

    def getMinimumSubsequenceLength(self) -> int:
        """
        get the minimum subsequence length
        :return: minimum subsequence length
        :rtype: int
        """
        return min(list(map(min,self.subSeqLengthList)))

    def getAverageItemPerSequenceLength(self) -> float:
        """
        get the average item length per sequence. It is sum of all item length divided by sequence length.
        :return: average item length per sequence
        :rtype: float
        """
        totalLength = sum(list(map(sum,self.subSeqLengthList)))
        return totalLength / len(self.database)

    def getMaximumSubsequenceLength(self) -> int:
        """
        get the maximum subsequence length
        :return: maximum subsequence length
        :rtype: int
        """
        return max(list(map(max,self.subSeqLengthList)))

    def getStandardDeviationSubsequenceLength(self) -> float:
        """
        get the standard deviation subsequence length
        :return: standard deviation subsequence length
        :rtype: float
        """
        allList=[]
        for i in self.subSeqLengthList:
            allList=allList+i
        return statistics.pstdev(allList)

    def getVarianceSubsequenceLength(self) -> float:
        """
        get the variance subSequence length
        :return: variance subSequence length
        :rtype: float
        """
        allList = []
        for i in self.subSeqLengthList:
            allList = allList + i
        return statistics.variance(allList)

    def getSortedListOfItemFrequencies(self) -> Dict[str, int]:
        """
        get sorted list of item frequencies
        :return: item frequencies
        :rtype: dict
        """
        itemFrequencies = {}
        for seq in self.database:
            for sub in self.database[seq]:
                for item in sub:
                    itemFrequencies[item] = itemFrequencies.get(item, 0)
                    itemFrequencies[item] += 1
        return {k: v for k, v in sorted(itemFrequencies.items(), key=lambda x: x[1], reverse=True)}

    def getFrequenciesInRange(self) -> Dict[int, int]:
        """
        get sorted list of item frequencies in some range
        :return: item separated by its frequencies
        :rtype: dict
        """
        fre = self.getSortedListOfItemFrequencies()
        rangeFrequencies = {}
        maximum = max([i for i in fre.values()])
        values = [int(i * maximum / 6) for i in range(1, 6)]
        va = len({key: val for key, val in fre.items() if val > 0 and val < values[0]})
        rangeFrequencies[values[0]] = va
        for i in range(1, len(values)):
            va = len({key: val for key, val in fre.items() if val < values[i] and val > values[i - 1]})
            rangeFrequencies[values[i]] = va
        return rangeFrequencies

    def getSequencialLengthDistribution(self) -> Dict[int, int]:
        """
        get Sequence length Distribution
        :return: Sequence length
        :rtype: dict
        """
        transactionLength = {}
        for length in self.seqLengthList:
            transactionLength[length] = transactionLength.get(length, 0)
            transactionLength[length] += 1
        return {k: v for k, v in sorted(transactionLength.items(), key=lambda x: x[0])}

    def getSubsequencialLengthDistribution(self) -> Dict[int, int]:
        """
        get subSequence length distribution
        :return: subSequence length
        :rtype: dict
        """
        transactionLength = {}
        for sublen in self.subSeqLengthList:
            for length in sublen:
                transactionLength[length] = transactionLength.get(length, 0)
                transactionLength[length] += 1
        return {k: v for k, v in sorted(transactionLength.items(), key=lambda x: x[0])}

    def run(self) -> None:
        self.readDatabase()

    def printStats(self) -> None:
        """
        To print the all status of sequence database
        """
        print(f'Database size (total no of sequence) : {self.getDatabaseSize()}')
        print(f'Number of items : {self.getTotalNumberOfItems()}')
        print(f'Minimum Sequence Size : {self.getMinimumSequenceLength()}')
        print(f'Average Sequence Size : {self.getAverageSubsequencePerSequenceLength()}')
        print(f'Maximum Sequence Size : {self.getMaximumSequenceLength()}')
        print(f'Standard Deviation Sequence Size : {self.getStandardDeviationSequenceLength()}')
        print(f'Variance in Sequence Sizes : {self.getVarianceSequenceLength()}')
        print(f'Sequence size (total no of subsequence) : {self.getSequenceSize()}')
        print(f'Minimum subSequence Size : {self.getMinimumSubsequenceLength()}')
        print(f'Average subSequence Size : {self.getAverageItemPerSubsequenceLength()}')
        print(f'Maximum subSequence Size : {self.getMaximumSubsequenceLength()}')
        print(f'Standard Deviation Sequence Size : {self.getStandardDeviationSubsequenceLength()}')
        print(f'Variance in Sequence Sizes : {self.getVarianceSubsequenceLength()}')

    def plotGraphs(self) -> None:
        """
        To plot the  distribution about items, subsequences in sequence and items in subsequence
        """
        itemFrequencies = self.getFrequenciesInRange()
        seqLen = self.getSequencialLengthDistribution()
        subLen=self.getSubsequencialLengthDistribution()
        plt.plotLineGraphFromDictionary(itemFrequencies, 100, 'Frequency', 'No of items', 'frequency')
        plt.plotLineGraphFromDictionary(seqLen, 100, 'sequence length', 'sequence length', 'frequency')
        plt.plotLineGraphFromDictionary(subLen, 100, 'subsequence length', 'subsequence length', 'frequency')

if __name__ == '__main__':
    _ap=str()
    if len(sys.argv)==3 or len(sys.argv)==2:
        if len(sys.argv)==3:
            _ap=sequentialDatabase(sys.argv[1],sys.argv[2])
        if len(sys.argv) == 2:
            _ap = sequentialDatabase(sys.argv[1])
        _ap.run()
        _ap.printStats()
        _ap.plotGraphs()
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



class ExtendedEdge:
    def __init__(self, v1, v2, vLabel1, vLabel2, edgeLabel):
        self.v1 = v1
        self.v2 = v2
        self.vLabel1 = vLabel1
        self.vLabel2 = vLabel2
        self.edgeLabel = edgeLabel
        self.hashcode = (1 + v1) * 100 + (1 + v2) * 50 + (1 + vLabel1) * 30 + (1 + vLabel2) * 20 + (1 + edgeLabel)

    def smallerThan(self, that):
        if that is None:
            return True

        x1, x2, y1, y2 = self.v1, self.v2, that.v1, that.v2

        if self.pairSmallerThan(x1, x2, y1, y2):
            return True
        elif x1 == y1 and x2 == y2:
            return (self.vLabel1 < that.vLabel1 or
                    (self.vLabel1 == that.vLabel1 and self.vLabel2 < that.vLabel2) or
                    (self.vLabel1 == that.vLabel1 and self.vLabel2 == that.vLabel2 and
                     self.edgeLabel < that.edgeLabel))
        else:
            return False

  
    def smallerThanOriginal(self, that):
        if that is None:
            return True

        x1, x2, y1, y2 = self.v1, self.v2, that.v1, that.v2

        if self.pairSmallerThan(x1, x2, y1, y2):
            return True
        elif x1 == y1 and x2 == y2:
            return (self.vLabel1 < that.vLabel1 or
                    (self.vLabel1 == that.vLabel1 and self.edgeLabel < that.edgeLabel) or
                    (self.vLabel1 == that.vLabel1 and self.edgeLabel == that.edgeLabel and
                     self.vLabel2 < that.vLabel2))
        else:
            return False

    def pairSmallerThan(self, x1, x2, y1, y2):
        xForward = x1 < x2
        yForward = y1 < y2

        if xForward and yForward:
            return x2 < y2 or (x2 == y2 and x1 > y1)
        elif not xForward and not yForward:
            return x1 < y1 or (x1 == y1 and x2 < y2)
        elif xForward:
            return x2 <= y1
        else:
            return x1 < y2

    def __hash__(self):
        return self.hashcode

    def __eq__(self, other):
        if not isinstance(other, ExtendedEdge):
            return False
        return (self.v1 == other.v1 and self.v2 == other.v2 and
                self.vLabel1 == other.vLabel1 and self.vLabel2 == other.vLabel2 and
                self.edgeLabel == other.edgeLabel)

    def __repr__(self):
        return f"<{self.v1},{self.v2},{self.vLabel1},{self.vLabel2},{self.edgeLabel}>"

    def getV1(self):
        return self.v1
    
    def getV2(self):
        return self.v2
    
    def getVLabel1(self):
        return self.vLabel1
    
    def getVLabel2(self):
        return self.vLabel2
    
    def getEdgeLabel(self):
        return self.edgeLabel



# temporalToFuzzy is used to convert the transactional database into Fuzzy transactional database.
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#             from PAMI.extras.FuzzyTransformation import temporalToFuzzy as db
#
#             obj = db.temporalToFuzzy(iFile, FuzFile, oFile, "\t" )
#
#             obj.startConvert()
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from PAMI.extras.fuzzyTransformation import abstract as _ab


class temporalToFuzzy(_ab._convert):
    """

    :Description:
            temporalToFuzzy is used to convert the temporal database into Fuzzy temporal database.


    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent patterns
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  fuzFile: str :
                   Name of the Fuzzy File to process set of data.
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    **Importing this algorithm into a python program**
    --------------------------------------------------------
    .. code-block:: python

            from PAMI.extras.FuzzyTransformation import temporalToFuzzy as db

            obj = db.temporalToFuzzy(iFile, FuzFile, oFile, "\t" )

            obj.startConvert()

    """

    _iFile: str = ' '
    _fuzFile: str = ' '
    _oFile: str = ' '


    def __init__(self, iFile: str, fuzFile: str, oFile: str, sep: str='\t'):
        self._iFile = iFile
        self._fuzFile = fuzFile
        self._oFile = oFile
        self._sep = sep
        self._RegionsCal = []
        self._RegionsLabel = []
        self._LabelKey = {}
        self._LabelKeyOne = {}
        self._dbLen = 0
        self._list = []
        self._transactionsDB = []
        self._fuzzyValuesDB = []
        self._tsDB = []
        self._fuzzyRegionReferenceMap = {}

    def _creatingItemSets(self) -> None:
        """
          To process the input file and store the timestamps, items, and their values as lists respectively.
        """
        self._transactionsDB, self._fuzzyValuesDB, self._tsDB = [], [], []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self._transactionsDB = self._iFile['Transactions'].tolist()
            if 'fuzzyValues' in i:
                self._fuzzyValuesDB = self._iFile['Utilities'].tolist()

        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line = line.decode("utf-8")
                    line = line.split("\n")[0]
                    parts = line.split(":")
                    items = parts[0].split(self._sep)
                    quantities = parts[1].split(self._sep)
                    self._tsDB.append(int(items[0]))
                    self._transactionsDB.append([x for x in items[1:]])
                    self._fuzzyValuesDB.append([x for x in quantities])
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            parts = line.split(":")
                            parts[0] = parts[0].strip()
                            parts[1] = parts[1].strip()
                            items = parts[0].split(self._sep)
                            quantities = parts[1].split(self._sep)
                            self._tsDB.append(int(items[0]))
                            self._transactionsDB.append([x for x in items[1:]])
                            self._fuzzyValuesDB.append([x for x in quantities])
                except IOError:
                    print("File Not Found")
                    quit()

    def _fuzzyMembershipFunc(self) -> None:
        """
        The Fuzzy file is processed and labels created according the boundaries specified in input file.
        """
        try:
            with open(self._fuzFile, 'r', encoding='utf-8') as f:
                count = 0
                for line in f:
                    line = line.split("\n")[0]
                    parts = line.split(self._sep)
                    lowerBound = parts[0].strip()
                    upperBound = parts[1].strip()
                    lb_Label = parts[2].strip()
                    ub_Label = parts[3].strip()
                    self._RegionsCal.append([int(lowerBound), int(upperBound)])
                    self._RegionsLabel.append([lb_Label, ub_Label])
                    for i in range(0, 2):
                        if ub_Label not in self._LabelKey:
                            self._LabelKey[ub_Label] = count
                            count += 1
            self._LabelKeyOne = {v: k for k, v in self._LabelKey.items()}
        except IOError:
            print("File Not Found")
            quit()

    def _Regions(self, quantity: int) -> None:
        """
        calculate the labelled region of input "quantity"
        :param quantity: represents the quantity of item
        :type quantity: int
        :return: None
        """
        self._list = [0] * len(self._LabelKey)
        if self._RegionsCal[0][0] < quantity <= self._RegionsCal[0][1]:
            self._list[0] = 1
            return
        elif quantity >= self._RegionsCal[-1][0]:
            self._list[-1] = 1
            return
        else:
            for i in range(1, len(self._RegionsCal) - 1):
                if self._RegionsCal[i][0] <= quantity <= self._RegionsCal[i][1]:
                    base = self._RegionsCal[i][1] - self._RegionsCal[i][0]
                    self._list[self._LabelKey[self._RegionsLabel[i][0]]] = float((self._RegionsCal[i][1] - quantity) / base)
                    self._list[self._LabelKey[self._RegionsLabel[i][1]]] = float((quantity - self._RegionsCal[i][0]) / base)
            return

    def startConvert(self) -> None:
        """
        Main method to convert the temporal database into fuzzy database.
        """
        _writer = open(self._oFile, 'w+')
        self._creatingItemSets()
        self._fuzzyMembershipFunc()
        for line in range(len(self._transactionsDB)):
            item_list = self._transactionsDB[line]
            fuzzyValues_list = self._fuzzyValuesDB[line]
            self._dbLen += 1
            s = str(self._tsDB[line])
            ss = str()
            for i in range(0, len(item_list)):
                item = item_list[i]
                fuzzy_ref = fuzzyValues_list[i]
                regionsList = self._Regions(int(fuzzy_ref))
                self._fuzzyRegionReferenceMap[fuzzy_ref] = regionsList
                s1 = [self._list.index(i) for i in self._list if i!=0]
                for k in s1:
                    s = s + item + '.' + self._LabelKeyOne[k] + '\t'
                    st = round(self._list[k], 2)
                    ss = ss + str(st) + '\t'
            s2 = s.strip('\t') + ":" + ss
            _writer.write("%s \n" % s2)



if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = temporalToFuzzy(_ab._sys.argv[1], _ab._sys.argv[2], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = temporalToFuzzy(_ab._sys.argv[1], _ab._sys.argv[2],  _ab._sys.argv[3])
        _ap.startConvert()
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")



class SparseTriangularMatrix:
    def __init__(self):
        self.matrix = {}  

    def __str__(self):
        temp = []
        for key in sorted(self.matrix.keys()):
            subkeys = self.matrix[key]
            subkeyStr = " ".join(f"{subkey}:{count}" for subkey, count in subkeys.items())
            temp.append(f"{key}: {subkeyStr}\n")
        return "".join(temp)

    def incrementCount(self, i, j):
        if i < j:
            key, subkey = i, j
        else:
            key, subkey = j, i

        if key not in self.matrix:
            self.matrix[key] = {subkey: 1}
        else:
            if subkey not in self.matrix[key]:
                self.matrix[key][subkey] = 1
            else:
                self.matrix[key][subkey] += 1

    def getSupportForItems(self, i, j):
        smaller, larger = min(i, j), max(i, j)
        return self.matrix.get(smaller, {}).get(larger, 0)

    def setSupport(self, i, j, support):
        smaller, larger = min(i, j), max(i, j)
        
        if smaller not in self.matrix:
            self.matrix[smaller] = {larger: support}
        else:
            self.matrix[smaller][larger] = support

    def removeInfrequentEntriesFromMatrix(self, minsup):
        for key in list(self.matrix.keys()):
            for subkey, value in list(self.matrix[key].items()):
                if value < minsup:
                    del self.matrix[key][subkey]
            if not self.matrix[key]:
                del self.matrix[key]



# SPADE is one of the fundamental algorithm to discover sequential frequent patterns in a transactional database.
# This program employs SPADE property (or downward closure property) to  reduce the search space effectively.
# This algorithm employs breadth-first search technique when 1-2 length patterns and depth-first search when above 3 length patterns to find the complete set of frequent patterns in a transactional database.
#
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#
#             import PAMI.sequentialPatternMining.basic.SPADE as alg
#
#             obj = alg.SPADE(iFile, minSup)
#
#             obj.startMine()
#
#             sequentialPatternMining = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#




__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""


import pandas as pd
from deprecated import deprecated

from PAMI.sequentialPatternMining.basic import abstract as _ab

_ab._sys.setrecursionlimit(10000)

class SPADE(_ab._sequentialPatterns):
    """
    :Description:

        * SPADE is one of the fundamental algorithm to discover sequential frequent patterns in a transactional database.
        * This program employs SPADE property (or downward closure property) to  reduce the search space effectively.
        * This algorithm employs breadth-first search technique when 1-2 length patterns and depth-first serch when above 3 length patterns to find the complete set of frequent patterns in a transactional database.

    :Reference:   Mohammed J. Zaki. 2001. SPADE: An Efficient Algorithm for Mining Frequent Sequences. Mach. Learn. 42, 1-2 (January 2001), 31-60. DOI=10.1023/A:1007652502315 http://dx.doi.org/10.1023/A:1007652502315

    :param  iFile: str :
                   Name of the Input file to mine complete set of  Sequential frequent patterns
    :param  oFile: str :
                   Name of the output file to store complete set of  Sequential frequent patterns
    :param  minSup: float or int or str :
                    minSup measure constraints the minimum number of transactions in a database where a pattern must appear
                    Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

            iFile : str
                Input file name or path of the input file
            oFile : str
                Name of the output file or the path of output file
            minSup: float or int or str
                The user can specify minSup either in count or proportion of database size.
                If the program detects the data type of minSup is integer, then it treats minSup is expressed in count.
                Otherwise, it will be treated as float.
                Example: minSup=10 will be treated as integer, while minSup=10.0 will be treated as float
            sep : str
                This variable is used to distinguish items from one another in a transaction. The default seperator is tab space or \t.
                However, the users can override their default separator.
            startTime:float
                To record the start time of the mining process
            endTime:float
                To record the completion time of the mining process
            finalPatterns: dict
                Storing the complete set of patterns in a dictionary variable
            memoryUSS : float
                To store the total amount of USS memory consumed by the program
            memoryRSS : float
                To store the total amount of RSS memory consumed by the program
            Database : list
                To store the transactions of a database in list
            _xLenDatabase: dict
                To store the datas in different sequence separated by sequence, rownumber, length.
            _xLenDatabaseSame : dict
                To store the datas in same sequence separated by sequence, rownumber, length.

    :Methods:

            startMine()
                Mining process will start from here
            getPatterns()
                Complete set of patterns will be retrieved with this function
            savePatterns(oFile)
                Complete set of frequent patterns will be loaded in to an output file
            getPatternsAsDataFrame()
                Complete set of frequent patterns will be loaded in to a dataframe
            getMemoryUSS()
                Total amount of USS memory consumed by the mining process will be retrieved from this function
            getMemoryRSS()
                Total amount of RSS memory consumed by the mining process will be retrieved from this function
            getRuntime()
                Total amount of runtime taken by the mining process will be retrieved from this function
            candidateToFrequent(candidateList)
                Generates frequent patterns from the candidate patterns
            frequentToCandidate(frequentList, length)
                Generates candidate patterns from the frequent patterns

    **Methods to execute code on terminal**
    -------------------------------------------
    .. code-block:: console


       Format:

       (.venv) $ python3 SPADE.py <inputFile> <outputFile> <minSup>

       Example usage:

       (.venv) $ python3 SPADE.py sampleDB.txt patterns.txt 10.0


               .. note:: minSup will be considered in times of minSup and count of database transactions

    **Importing this algorithm into a python program**
    ----------------------------------------------------
    .. code-block:: python

            import PAMI.sequentialPatternMining.basic.SPADE as alg

            obj = alg.SPADE(iFile, minSup)

            obj.startMine()

            sequentialPatternMining = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    ---------------

              The complete program was written by Suzuki Shota under the supervision of Professor Rage  Uday Kiran.

    """

    _minSup = float()
    _startTime = float()
    _endTime = float()
    _finalPatterns = {}
    _iFile = " "
    _oFile = " "
    _sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _Database = []
    _xLenDatabase={}
    _xLenDatabaseSame = {}
    def _creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self._Database = []

        if isinstance(self._iFile, _ab._pd.DataFrame):
            temp = []
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                temp = self._iFile['Transactions'].tolist()
            if "tid" in i:
                temp2=self._iFile[''].tolist()
            addList=[]
            addList.append(temp[0])
            for k in range(len(temp)-1):
                if temp2[k]==temp[k+1]:
                    addList.append(temp[k+1])
                else:
                    self._Database.append(addList)
                    addList=[]
                    addList.append(temp[k+1])
            self._Database.append(addList)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    temp.pop()
                    self._Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split('-1')]
                            temp = [x for x in temp if x ]
                            temp.pop()

                            seq = []
                            for i in temp:
                                k = -2
                                if len(i)>1:
                                    seq.append(list(sorted(set(i.split()))))

                                else:
                                    seq.append(i)

                            self._Database.append(seq)

                except IOError:
                    print("File Not Found")
                    quit()

    def _convert(self, value):
        """
        To convert the user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self._Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self._Database) * value)
            else:
                value = int(value)
        return value


    def make1LenDatabase(self):
        """
        To make 1 length frequent patterns by breadth-first search technique   and update Database to sequential database
        """

        idDatabase={}
        alreadyInData=[]
        lineNumber=0
        alreadyInLine={}
        for line in self._Database:
            seqNumber=0
            for seq in line:

                for data in seq:
                    if data in alreadyInData:
                        if lineNumber in alreadyInLine[data]:
                            idDatabase[data][lineNumber].append(seqNumber)
                        else:
                            idDatabase[data][lineNumber] = []
                            idDatabase[data][lineNumber].append(seqNumber)
                            alreadyInLine[data].append(lineNumber)
                    else:
                        idDatabase[data]={}
                        idDatabase[data][lineNumber]=[]
                        idDatabase[data][lineNumber].append(seqNumber)
                        alreadyInData.append(data)
                        alreadyInLine[data]=[]
                        alreadyInLine[data].append(lineNumber)



                seqNumber+=1
            lineNumber+=1

        newDatabase={i :idDatabase[i] for i in idDatabase.keys()}
        for key in idDatabase.keys():
            if len(idDatabase[key].keys())<self._minSup:
                newDatabase.pop(key)
            else:
                self._finalPatterns[str(key)]=len(idDatabase[key].keys())
        self._Database=newDatabase


    def make2LenDatabase(self):
        """
        To make 2 length frequent patterns by joining two one length patterns by breadth-first search technique  and update xlen Database to sequential database
        """
        self._xLenDatabase = {}
        keyList=[i for i in self._Database.keys()]
        nextDatabase={i:{} for i in self._Database.keys()}
        nextDatabaseSame={i:{} for i in self._Database.keys()}
        keyNumber=-1
        for key1 in keyList:
            keyNumber+=1
            for key2 in keyList[keyNumber:]:
                if key1!=key2:


                    if len(self._Database[key1].keys())>=len(self._Database[key1].keys()):
                        nextDatabase[key1][key2] = {}
                        nextDatabase[key2][key1] = {}
                        nextDatabaseSame[key1][key2] = {}

                        for seq in self._Database[key2].keys():
                            if seq in self._Database[key1].keys():
                                x=[i for i in self._Database[key1][seq] if i >self._Database[key2][seq][0]]
                                if len(x) != 0:
                                    nextDatabase[key2][key1][seq]=x
                                x = [i for i in self._Database[key2][seq] if i > self._Database[key1][seq][0]]
                                if len(x) != 0:
                                    nextDatabase[key1][key2][seq]=x
                                x=list(sorted(set(self._Database[key1][seq])&set(self._Database[key2][seq])))
                                if len(x)!=0:
                                    nextDatabaseSame[key1][key2][seq]=x
                    else:
                        nextDatabase[key1][key2] = {}
                        nextDatabase[key2][key1] = {}
                        nextDatabaseSame[key1][key2] = {}

                        for seq in self._Database[key1].keys():
                            x = [i for i in self._Database[key1][seq] if
                                                                    i > self._Database[key2][seq][0]]
                            if len(x)!=0:
                                nextDatabase[key2][key1][seq]=0
                            x = [i for i in self._Database[key2][seq] if
                                                                    i > self._Database[key1][seq][0]]
                            if len(x)!=0:
                                nextDatabase[key1][key2][seq]=x
                            x= list(
                                sorted(set(self._Database[key1][seq]) & set(self._Database[key2][seq])))
                            if len(x)!=0:
                                nextDatabaseSame[key1][key2][seq]=x
                else:
                    nextDatabase[key1][key2] = {}
                    for seq in self._Database[key2].keys():
                        if len(self._Database[key1][seq])>=2:
                            nextDatabase[key1][key2][seq]= self._Database[key2][seq][1:]
        self._xLenDatabase[2] = {tuple([i]): {} for i in nextDatabase.keys()}
        for key1 in nextDatabase.keys():
            for key2 in nextDatabase[key1].keys():
                if len(nextDatabase[key1][key2].keys())>=self._minSup:
                    self._finalPatterns[str((key1,-1,key2,-1))]=len(nextDatabase[key1][key2].keys())
                    self._xLenDatabase[2][tuple([key1])][key2]=nextDatabase[key1][key2]
        self._xLenDatabaseSame[2]={tuple([i]): {} for i in nextDatabaseSame.keys()}
        for key1 in nextDatabaseSame.keys():
            for key2 in nextDatabaseSame[key1].keys():
                if len(nextDatabaseSame[key1][key2].keys()) >= self._minSup:
                    self._finalPatterns[str((key1,key2,-1))]=len(nextDatabaseSame[key1][key2].keys())
                    self._xLenDatabaseSame[2][tuple([key1])][key2]={i:nextDatabaseSame[key1][key2][i] for i in nextDatabaseSame[key1][key2].keys()}
                    self._xLenDatabaseSame[2][tuple([key2])][key1] = {i: nextDatabaseSame[key1][key2][i] for i in nextDatabaseSame[key1][key2].keys()}

    def make3LenDatabase(self):
        """
        To call each 2 length patterns to make 3 length frequent patterns depth-first search technique
        """
        for i in self._xLenDatabase[2].keys():
            for k in self._xLenDatabase[2][i].keys():
                self.makexLenDatabase(2,i,k)
        for i in self._xLenDatabaseSame[2].keys():
            for k in self._xLenDatabaseSame[2][i].keys():
                self.makexLenDatabaseSame(2,i,k)

    def makexLenDatabase(self, rowLen, bs, latestWord):
            """
            To make "rowLen" length frequent patterns from pattern which the latest word is in same seq  by joining "rowLen"-1 length patterns by depth-first search technique  and update xlenDatabase to sequential database

            :param rowLen: row length of patterns.
            :param bs : patterns without the latest one
            :param latestWord : latest word of patterns
            """
            if rowLen+1 not in self._xLenDatabase:
                self._xLenDatabase[rowLen+1]={}
                self._xLenDatabaseSame[rowLen+1]={}
            for latestWord2 in self._xLenDatabase[rowLen][bs].keys():
                if latestWord != latestWord2:

                    if len(self._xLenDatabase[rowLen][bs][latestWord].keys()) <= len(self._xLenDatabase[rowLen][bs][latestWord2].keys()):
                        next={}
                        next2={}
                        nextSame={}

                        for seq in self._xLenDatabase[rowLen][bs][latestWord].keys():
                            if seq in self._xLenDatabase[rowLen][bs][latestWord2].keys():
                                if self._xLenDatabase[rowLen][bs][latestWord2][seq]!=[] and self._xLenDatabase[rowLen][bs][latestWord][seq]!=[]:
                                    x = [i for i in self._xLenDatabase[rowLen][bs][latestWord2][seq] if
                                            i > self._xLenDatabase[rowLen][bs][latestWord][seq][0]]
                                    if len(x) != 0:
                                        next[seq] = x
                                    x = [i for i in self._xLenDatabase[rowLen][bs][latestWord][seq] if
                                            i > self._xLenDatabase[rowLen][bs][latestWord2][seq][0]]
                                    if len(x) != 0:
                                        next2[seq] = x
                                    x = list(sorted(set(self._xLenDatabase[rowLen][bs][latestWord][seq]) & set(
                                        self._xLenDatabase[rowLen][bs][latestWord2][seq])))
                                    if len(x) != 0:
                                        nextSame[seq] = x
                        if len(next)>=self._minSup:
                            nextRow,nextbs= self.makeNextRow(bs,latestWord,latestWord2)
                            if str(nextRow) not in self._finalPatterns.keys():
                                self._finalPatterns[str(nextRow)] = len(next)
                                if nextbs not in self._xLenDatabase[rowLen + 1]:
                                    self._xLenDatabase[rowLen + 1][nextbs]={}
                                self._xLenDatabase[rowLen+1][nextbs][latestWord2]={i:next[i] for i in next }
                                self.makexLenDatabase(rowLen+1,nextbs,latestWord2)
                        if len(next2)>=self._minSup:
                            nextRow,nextbs = self.makeNextRow(bs, latestWord2, latestWord)
                            if str(nextRow) not in self._finalPatterns.keys():
                                if nextbs not in self._xLenDatabase[rowLen + 1]:
                                    self._xLenDatabase[rowLen + 1][nextbs]={}
                                self._finalPatterns[str(nextRow)] = len(next2)
                                self._xLenDatabase[rowLen+1][nextbs][latestWord] = {i:next2[i] for i in next2 }
                                self.makexLenDatabase(rowLen+1, nextbs, latestWord)
                        if len(nextSame) >= self._minSup:
                            nextRow,nextbs ,nextlste= self.makeNextRowSame3(bs, latestWord, latestWord2)
                            if str(nextRow) not in self._finalPatterns.keys():
                                if nextbs not in self._xLenDatabaseSame[rowLen + 1]:
                                    self._xLenDatabaseSame[rowLen + 1][nextbs]={}
                                self._finalPatterns[str(nextRow)] = len(nextSame)
                                self._xLenDatabaseSame[rowLen+1][nextbs][nextlste]={i:nextSame[i] for i in nextSame }
                                self.makexLenDatabaseSame(rowLen+1, nextbs, nextlste)

                    else:
                        next = {}
                        next2 = {}
                        nextSame = {}

                        for seq in self._xLenDatabase[rowLen][bs][latestWord2].keys():
                            if seq in self._xLenDatabase[rowLen][bs][latestWord].keys():
                                if self._xLenDatabase[rowLen][bs][latestWord2][seq] != [] and self._xLenDatabase[rowLen][bs][latestWord][seq] != []:
                                    x = [i for i in self._xLenDatabase[rowLen][bs][latestWord2][seq] if i > self._xLenDatabase[rowLen][bs][latestWord][seq][0]]
                                    if len(x)!=0:
                                        next[seq]=x
                                    x= [i for i in self._xLenDatabase[rowLen][bs][latestWord][seq] if i > self._xLenDatabase[rowLen][bs][latestWord2][seq][0]]
                                    if len(x)!=0:
                                        next2[seq]=x
                                    x= list(sorted(set(self._xLenDatabase[rowLen][bs][latestWord][seq]) & set(
                                            self._xLenDatabase[rowLen][bs][latestWord2][seq])))
                                    if len(x)!=0:
                                        nextSame[seq]=x
                        if len(next) >= self._minSup:
                            nextRow,nextbs = self.makeNextRow(bs, latestWord, latestWord2)
                            if str(nextRow) not in self._finalPatterns.keys():
                                if nextbs not in self._xLenDatabase[rowLen + 1]:
                                    self._xLenDatabase[rowLen+1][nextbs]={}
                                self._finalPatterns[str(nextRow)] = len(next)
                                self._xLenDatabase[rowLen + 1][nextbs][latestWord2] ={i:next[i] for i in next }
                                self.makexLenDatabase(rowLen+1, nextbs, latestWord2)
                        if len(next2) >= self._minSup:
                            nextRow,nextbs = self.makeNextRow(bs, latestWord2, latestWord)
                            if str(nextRow) not in self._finalPatterns.keys():
                                if nextbs not in self._xLenDatabase[rowLen + 1]:
                                    self._xLenDatabase[rowLen + 1][nextbs]={}
                                self._finalPatterns[str(nextRow)] = len(next2)
                                self._xLenDatabase[rowLen+1][nextbs][latestWord] = {i:next2[i] for i in next2 }
                                self.makexLenDatabase(rowLen+1, nextbs, latestWord)
                        if len(nextSame) >= self._minSup:
                            nextRow,nextbs,nextlate = self.makeNextRowSame3(bs,latestWord,latestWord2)
                            if str(nextRow) not in self._finalPatterns.keys():
                                if nextbs not in self._xLenDatabaseSame[rowLen + 1]:
                                    self._xLenDatabaseSame[rowLen + 1][nextbs]={}
                                self._finalPatterns[str(nextRow)] = len(nextSame)
                                self._xLenDatabaseSame[rowLen+1][nextbs][nextlate] = {i:nextSame[i] for i in nextSame}
                                self.makexLenDatabaseSame(rowLen+1, nextbs, nextlate)

                else:
                    next= {}
                    for seq in self._xLenDatabase[rowLen][bs][latestWord2].keys():
                        if len(self._xLenDatabase[rowLen][bs][latestWord][seq])>=2:
                            next[seq]= self._xLenDatabase[rowLen][bs][latestWord][seq][1:]
                    if len(next) >= self._minSup:
                        nextRow, nextbs= self.makeNextRow(bs,latestWord,latestWord2)
                        if str(nextRow) not in self._finalPatterns.keys():
                            if nextbs not in self._xLenDatabase[rowLen+1]:
                                self._xLenDatabase[rowLen+1][nextbs]={}
                            self._finalPatterns[str(nextRow)] = len(next)
                            self._xLenDatabase[rowLen+1][nextbs][latestWord2] ={i:next[i] for i in next }
                            self.makexLenDatabase(rowLen+1, nextbs, latestWord2)
            if bs in self._xLenDatabaseSame[rowLen]:
                for latestWord2 in self._xLenDatabaseSame[rowLen][bs]:


                        if len(self._xLenDatabase[rowLen][bs][latestWord].keys()) <= len(
                                self._xLenDatabaseSame[rowLen][bs][latestWord2].keys()):
                            next = {}

                            for seq in self._xLenDatabase[rowLen][bs][latestWord].keys():
                                if seq in self._xLenDatabaseSame[rowLen][bs][latestWord2].keys():
                                    if self._xLenDatabaseSame[rowLen][bs][latestWord2][seq] != []:
                                        x= [i for i in self._xLenDatabase[rowLen][bs][latestWord][seq] if i > self._xLenDatabaseSame[rowLen][bs][latestWord2][seq][0]]
                                        if len(x) != 0:
                                            next[seq] = x
                            if len(next) >= self._minSup:

                                nextRow ,nextbs= self.makeNextRowSame(bs, latestWord2, latestWord)
                                if str(nextRow) not in self._finalPatterns.keys():
                                    if nextbs not in self._xLenDatabase[rowLen + 1]:
                                        self._xLenDatabase[rowLen + 1][nextbs] = {}
                                    self._finalPatterns[str(nextRow)] = len(next)
                                    self._xLenDatabase[rowLen + 1][nextbs][latestWord] = {i:next[i] for i in next }
                                    self.makexLenDatabase(rowLen + 1, nextbs, latestWord)

                        else:
                            next = {}
                            for seq in self._xLenDatabaseSame[rowLen][bs][latestWord2].keys():
                                if seq in self._xLenDatabase[rowLen][bs][latestWord].keys():
                                    if self._xLenDatabaseSame[rowLen][bs][latestWord2][seq] != [] :
                                        x= [i for i in self._xLenDatabase[rowLen][bs][latestWord][seq] if
                                                         i > self._xLenDatabaseSame[rowLen][bs][latestWord2][seq][0]]
                                        if len(x) != 0:
                                            next[seq] = x
                            if len(next) >= self._minSup:
                                nextRow,nextbs = self.makeNextRowSame(bs, latestWord2, latestWord)
                                if str(nextRow) not in self._finalPatterns.keys():
                                    if nextbs not in self._xLenDatabase[rowLen + 1]:
                                        self._xLenDatabase[rowLen + 1][nextbs] = {}
                                    self._finalPatterns[str(nextRow)] = len(next)
                                    self._xLenDatabase[rowLen+1][nextbs][latestWord] = {i:next[i] for i in next }
                                    self.makexLenDatabase(rowLen + 1, nextbs,latestWord)

    def makexLenDatabaseSame(self, rowLen, bs, latestWord):
        """
        To make 3 or more length frequent patterns from pattern which the latest word is in different seq  by depth-first search technique  and update xlenDatabase to sequential database

        :param rowLen: row length of previous patterns.
        :param bs : previous patterns without the latest one
        :param latestWord : latest word of previous patterns
        """
        if rowLen + 1 not in self._xLenDatabase:
            self._xLenDatabase[rowLen + 1] = {}
            self._xLenDatabaseSame[rowLen + 1] = {}
        if bs in self._xLenDatabase[rowLen]:
            for latestWord2 in self._xLenDatabase[rowLen][bs]:
                if len(self._xLenDatabaseSame[rowLen][bs][latestWord].keys()) <= len(self._xLenDatabase[rowLen][bs][latestWord2].keys()):
                    next = {}

                    for seq in self._xLenDatabaseSame[rowLen][bs][latestWord].keys():
                        if seq in self._xLenDatabase[rowLen][bs][latestWord2].keys():
                            if self._xLenDatabaseSame[rowLen][bs][latestWord][seq] != []:
                                x= [i for i in self._xLenDatabase[rowLen][bs][latestWord2][seq] if
                                             i > self._xLenDatabaseSame[rowLen][bs][latestWord][seq][0]]
                                if len(x) != 0:
                                    next[seq] = x
                    if len(next) >= self._minSup:
                        nextRow ,nextbs= self.makeNextRowSame(bs, latestWord, latestWord2)
                        if str(nextRow) not in self._finalPatterns.keys():
                            if nextbs not in self._xLenDatabase[rowLen + 1]:
                                self._xLenDatabase[rowLen + 1][nextbs] = {}
                            self._finalPatterns[str(nextRow)] = len(next)
                            self._xLenDatabase[rowLen + 1][nextbs][latestWord2]= {i:next[i] for i in next}
                            self.makexLenDatabase(rowLen + 1, nextbs, latestWord2)

                else:
                    next = {}

                    for seq in self._xLenDatabase[rowLen][bs][latestWord2].keys():
                        if seq in self._xLenDatabaseSame[rowLen][bs][latestWord].keys():
                            if self._xLenDatabaseSame[rowLen][bs][latestWord][seq] != []:
                                x= [i for i in self._xLenDatabase[rowLen][bs][latestWord2][seq] if
                                             i > self._xLenDatabaseSame[rowLen][bs][latestWord][seq][0]]
                                if len(x) != 0:
                                    next[seq] = x
                    if len(next) >= self._minSup:
                        nextRow,nextbs = self.makeNextRowSame(bs, latestWord, latestWord2)
                        if str(nextRow) not in self._finalPatterns.keys():
                            if nextbs not in self._xLenDatabase[rowLen + 1]:
                                self._xLenDatabase[rowLen + 1][nextbs] = {}
                            self._finalPatterns[str(nextRow)] = len(next)
                            self._xLenDatabase[rowLen + 1][nextbs][latestWord2] = {i:next[i] for i in next}
                            self.makexLenDatabase(rowLen + 1,nextbs, latestWord2)
        if bs in self._xLenDatabaseSame[rowLen]:
            for latestWord2 in self._xLenDatabaseSame[rowLen][bs]:
                if latestWord2!=latestWord:
                    if len(self._xLenDatabaseSame[rowLen][bs][latestWord].keys()) <= len(
                            self._xLenDatabaseSame[rowLen][bs][latestWord2].keys()):
                        next = {}

                        for seq in self._xLenDatabaseSame[rowLen][bs][latestWord].keys():
                            if seq in self._xLenDatabaseSame[rowLen][bs][latestWord2].keys():
                                x= list(sorted(set(self._xLenDatabaseSame[rowLen][bs][latestWord][seq]) & set(
                                        self._xLenDatabaseSame[rowLen][bs][latestWord2][seq])))
                                if len(x) != 0:
                                    next[seq] = x
                        if len(next) >= self._minSup:

                                nextRow, nextbs,nextLate= self.makeNextRowSame2(bs, latestWord, latestWord2)
                                if str(nextRow) not in self._finalPatterns.keys():
                                    if nextbs not in self._xLenDatabaseSame[rowLen+1]:
                                        self._xLenDatabaseSame[rowLen + 1][nextbs] = {}

                                    self._finalPatterns[str(nextRow)] = len(next)
                                    self._xLenDatabaseSame[rowLen + 1][nextbs][nextLate] = {i:next[i] for i in next}
                                    self.makexLenDatabaseSame(rowLen + 1, nextbs, nextLate)
                    else:
                        next = {}

                        for seq in self._xLenDatabaseSame[rowLen][bs][latestWord2].keys():
                            if seq in self._xLenDatabaseSame[rowLen][bs][latestWord].keys():
                                x= list(sorted(set(self._xLenDatabaseSame[rowLen][bs][latestWord][seq]) & set(
                                        self._xLenDatabaseSame[rowLen][bs][latestWord2][seq])))
                                if len(x) != 0:
                                    next[seq] = x
                        if len(next) >= self._minSup:

                                nextRow, nextbs,nextLate= self.makeNextRowSame2(bs, latestWord, latestWord2)
                                if str(nextRow) not in self._finalPatterns.keys():
                                    if nextbs not in self._xLenDatabaseSame[rowLen+1]:
                                        self._xLenDatabaseSame[rowLen + 1][nextbs] = {}

                                    self._finalPatterns[str(nextRow)] = len(next)
                                    self._xLenDatabaseSame[rowLen + 1][nextbs][nextLate] = {i:next[i] for i in next}
                                    self.makexLenDatabaseSame(rowLen + 1, nextbs, nextLate)

    def makeNextRow(self,bs, latestWord, latestWord2):
        """
        To make pattern row when two patterns have the latest word in different sequence

        :param bs : previous pattern without the latest one
        :param latestWord : latest word of one previous pattern
        :param latestWord2 : latest word of other previous pattern
        """

        bs=bs+(-1,latestWord)
        bs2=bs+(-1,latestWord2,-1)
        return bs2,bs

    def makeNextRowSame(self,bs, latestWord, latestWord2):
        """
        To make pattern row when one pattern have the latestWord1 in different sequence and other(latestWord2) in same

        :param bs : previous pattern without the latest one
        :param latestWord : latest word of one previous pattern in same sequence
        :param latestWord2 : latest word of other previous pattern in different sequence
        """

        bs=list(bs)
        x=1
        x2=[latestWord,]
        while bs:
            x=bs.pop()
            if x!=-1:
                x2.append(x)
            else:
                break
        x2=list(sorted(set(x2)))
        if len(bs)!=0:
            bs=tuple(bs)+(-1,)+tuple(x2)
        else:
            bs=tuple(x2)
        bs2=tuple(bs)+(-1,latestWord2,-1)
        return bs2,bs


    def makeNextRowSame2(self,bs, latestWord, latestWord2):
        """
        To make pattern row when two patterns have the latest word in same sequence

        :param bs : previous pattern without the latest one
        :param latestWord : latest word of one previous pattern
        :param latestWord2 : latest word of the other previous pattern
        """

        bs = list(bs)
        x = 1
        x2 = [latestWord, latestWord2]
        while bs:
            x = bs.pop()
            if x != -1:
                x2.append(x)
            else:
                break
        x2 = list(sorted(set(x2)))
        x3 = x2.pop()
        if len(bs)!=0:
            bs = tuple(bs)+(-1,)+ tuple(x2)
        else:
            bs = tuple(x2)
        bs2 = tuple(bs) + (x3, -1)

        return bs2, bs, x3


    def makeNextRowSame3(self, bs, latestWord, latestWord2):
        """
        To make pattern row when two patterns have the latest word in different sequence and both latest word is in same sequence

        :param bs : previous pattern without the latest one
        :param latestWord : latest word of one previous pattern
        :param latestWord2 : latest word of other previous pattern
        """

        x = list(sorted({latestWord, latestWord2}))
        x2 = x.pop()
        x3=x.pop()
        bs = bs + (-1,x3)
        bs2 = bs + (x2,)
        return  bs2,bs,x2

    @deprecated("It is recommended to use mine() instead of startMine() for mining process")
    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """
        self._Database = []
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self.make1LenDatabase()
        self.make2LenDatabase()
        self.make3LenDatabase()
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Sequential Frequent patterns were generated successfully using SPADE algorithm ")

    def Mine(self):
        """
        Frequent pattern mining process will start from here
        """
        self._Database = []
        self._startTime = _ab._time.time()
        self._creatingItemSets()
        self._minSup = self._convert(self._minSup)
        self.make1LenDatabase()
        self.make2LenDatabase()
        self.make3LenDatabase()
        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Sequential Frequent patterns were generated successfully using SPADE algorithm ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csv file
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            s1 = x + ":" + str(y)
            writer.write("%s \n" % s1)

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def printResults(self):
        """
        This function is used to prnt the results
        """
        print("Total number of Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = SPADE(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = SPADE(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        print("Total number of Frequent Patterns:", len(_Patterns))
        _ap.savePatterns(_ab._sys.argv[2])
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
    else:
        _ap = SPADE('text3.txt' ,80, '\t')
        _ap.startMine()
        _Patterns = _ap.getPatterns()
        _memUSS = _ap.getMemoryUSS()
        print("Total Memory in USS:", _memUSS)
        _memRSS = _ap.getMemoryRSS()
        print("Total Memory in RSS", _memRSS)
        _run = _ap.getRuntime()
        print("Total ExecutionTime in ms:", _run)
        print("Total number of Frequent Patterns:", len(_Patterns))
        print("Error! The number of input parameters do not match the total number of parameters provided")



import pickle

class DfsCode:
    def __init__(self):
        self.rightMost = -1
        self.size = 0
        self.rightMostPath = []  
        self.eeList = []

    def copy(self):
        return pickle.loads(pickle.dumps(self))

    def notPreOfRm(self, v):
        if len(self.rightMostPath) <= 1:
            return True
        return v != self.rightMostPath[-2]

    def getAllVLabels(self):
        labels = []
        vertexMap = {}
        for ee in self.eeList:
            v1, v1Label = ee.getV1(), ee.getVLabel1()
            v2, v2Label = ee.getV2(), ee.getVLabel2()
            vertexMap[v1] = v1Label
            vertexMap[v2] = v2Label
        
        count = 0
        while count in vertexMap:
            labels.append(vertexMap[count])
            count += 1
        return labels

    def add(self, ee):
        if self.size == 0:
            self.rightMost = 1
            self.rightMostPath.extend([0, 1])
        else:
            v1, v2 = ee.getV1(), ee.getV2()
            if v1 < v2:
                self.rightMost = v2
                while self.rightMostPath and self.rightMostPath[-1] > v1:
                    self.rightMostPath.pop()
                self.rightMostPath.append(v2)

        self.eeList.append(ee)
        self.size += 1

    def getAt(self, index):
        return self.eeList[index]

    def onRightMostPath(self, v):
        return v in self.rightMostPath

    def containEdge(self, v1, v2):
        for ee in self.eeList:
            if (ee.getV1() == v1 and ee.getV2() == v2) or (ee.getV1() == v2 and ee.getV2() == v1):
                return True
        return False

    def isEmpty(self):
        return not self.eeList

    def getRightMost(self):
        return self.rightMost

    def getRightMostPath(self):
        return self.rightMostPath

    def getEeList(self):
        return self.eeList

    def __str__(self):
        return "DfsCode: " + " ".join(str(ee) for ee in self.eeList)



class FrequentSubgraph:
    def __init__(self, dfsCode, setOfGraphsIds, support):
        self.dfsCode = dfsCode
        self.setOfGraphsIds = setOfGraphsIds
        self.support = support

    def __eq__(self, other):
        if not isinstance(other, FrequentSubgraph):
            return NotImplemented
        return self.support == other.support

    def __lt__(self, other):
        if not isinstance(other, FrequentSubgraph):
            return NotImplemented
        return self.support < other.support

    def __gt__(self, other):
        if not isinstance(other, FrequentSubgraph):
            return NotImplemented
        return self.support > other.support



import random
import pandas as pd


class generateUtilityTransactional:
    __transactionSize: int
    __numOfItems: int
    __avgTransactionLength: int
    __minUtilityValue: int
    __maxUtilityValue: int
    __minNumOfTimesAnItem: int
    __maxNumOfTimesAnItem: int
    __transactions: list[list[int]]
    __internalUtility: dict[str, list[int]]
    __externalUtility: list[list[int]]

    def __init__(self, transactionSize: int, numOfItems: int, avgTransactionLength: int,
                 minUtilityValue: int, maxUtilityValue: int,
                 minNumOfTimesAnItem: int, maxNumOfTimesAnItem: int) -> None:
        self.__transactionSize = transactionSize
        self.__numOfItems = numOfItems
        self.__avgTransactionLength = avgTransactionLength
        self.__minUtilityValue = minUtilityValue
        self.__maxUtilityValue = maxUtilityValue
        self.__minNumOfTimesAnItem = minNumOfTimesAnItem
        self.__maxNumOfTimesAnItem = maxNumOfTimesAnItem

        self.__transactions = list()
        self.__internalUtility = dict()
        self.__externalUtility = list()

    def generate(self) -> None:
        items = [i+1 for i in range(self.__numOfItems)]
        self.__transactions = [random.sample(items, random.randint(
            1, self.__avgTransactionLength*2)) for _ in range(self.__transactionSize)]
        self.__generateInternalUtility()
        self.__generateExternalUtility()

    def __generateInternalUtility(self) -> None:
        items = [i+1 for i in range(self.__numOfItems)]
        utilityValues = [random.randint(
            self.__minUtilityValue, self.__maxUtilityValue) for i in range(self.__numOfItems)]
        self.__internalUtility = {
            "items": items, "utilityValues": utilityValues}

    def __generateExternalUtility(self) -> None:
        self.__externalUtility = [[random.randint(self.__minNumOfTimesAnItem, self.__maxNumOfTimesAnItem) for _ in range(
            len(transaction))] for transaction in self.__transactions]

    def save(self, outputFile: str, sep="\t", type="utility") -> None:
        if (type == "utility"):
            with open(outputFile, 'w') as f:
                for transaction, exUtils in zip(self.__transactions, self.__externalUtility):
                    f.write(f"{sep.join(map(str, transaction))}:")
                    utilityValues = [
                        eu*self.__internalUtility["utilityValues"][item-1] for item, eu in zip(transaction, exUtils)]
                    f.write(
                        f"{sum(utilityValues)}:{sep.join(map(str, utilityValues))}\n")

        elif (type == "internal"):
            with open(outputFile, "w") as f:
                for item, utility in zip(self.__internalUtility["items"], self.__internalUtility["utilityValues"]):
                    f.write(f"{item}{sep}{utility}\n")

        elif (type == "external"):
            with open(outputFile, "w") as f:
                for transaction, exUtils in zip(self.__transactions, self.__externalUtility):
                    utils = list()
                    count = 0
                    for item in [i+1 for i in range(self.__numOfItems)]:
                        if item in transaction:
                            utils.append(exUtils[count])
                            count += 1
                        else:
                            utils.append(0)
                    f.write(f"{sep.join(map(str,utils))}\n")


if __name__ == "__main__":
    obj = generateUtilityTransactional(10, 10, 5, 10, 100, 1, 10)
    obj.generate()
    obj.save("transactionalUtility_test.csv", type="external")



import threading

class DfsThread(threading.Thread):
    def __init__(self, graphDb, candidates, minSup, tkgInstance):
        threading.Thread.__init__(self)
        self.graphDb = graphDb
        self.candidates = candidates
        self.minSup = minSup
        self.tkgInstance = tkgInstance

    def run(self):
        while not self.candidates.empty():
            _, candidate = self.candidates.get()
            if len(candidate.setOfGraphsIds) < self.minSup:
                break
            self.tkgInstance.gspanDynamicDFS(candidate.dfsCode, self.graphDb, candidate.setOfGraphsIds)



# cudaAprioriGCT is one of the fundamental algorithm to discover frequent patterns using CUDA in a transactional database. This program employs apriori property (or downward closure property) to  reduce the search space effectively. This algorithm employs breadth-first search technique to find the complete set of frequent patterns in a transactional database.
#
#
# **Importing this algorithm into a python program**
# ----------------------------------------------------
#
#             import PAMI.frequentPattern.cuda.cudaAprioriGCT as alg
#
#             obj = alg.cuAprioriGCT(iFile, minSup)
#
#             obj.mine()
#
#             frequentPatterns = obj.getPatterns()
#
#             print("Total number of Frequent Patterns:", len(frequentPatterns))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternInDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#




__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from deprecated import deprecated
from PAMI.frequentPattern.basic import abstract as _ab
# import abstract as _ab

import os
import time
import numpy as np
import pycuda.gpuarray as gpuarray
import psutil


class cudaAprioriGCT(_ab._frequentPatterns):
    """
    :Description: cudaAprioriGCT is one of the fundamental algorithm to discover frequent patterns in a transactional database. This program employs apriori property (or downward closure property) to  reduce the search space effectively. This algorithm employs breadth-first search technique to find the complete set of frequent patterns in a transactional database.

    :Reference:  Agrawal, R., Imieli ́nski, T., Swami, A.: Mining association rules between sets of items in large databases.
                In: SIGMOD. pp. 207–216 (1993), https://doi.org/10.1145/170035.170072

    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent patterns
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minSup: int :
                    The user can specify minSup either in count or proportion of database size. If the program detects the data type of minSup is integer, then it treats minSup is expressed in count. Otherwise, it will be treated as float.
    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        startTime : float
              To record the start time of the mining process

        endTime : float
              To record the completion time of the mining process

        finalPatterns : dict
              Storing the complete set of patterns in a dictionary variable

        memoryUSS : float
              To store the total amount of USS memory consumed by the program

        memoryRSS : float
              To store the total amount of RSS memory consumed by the program

        Database : list
              To store the transactions of a database in list



    **Methods to execute code on terminal**
    ----------------------------------------------------

    .. code-block:: console

      Format:

      (.venv) $ python3 cudaAprioriGCT.py <inputFile> <outputFile> <minSup>

      Example Usage:

      (.venv) $ python3 cudaAprioriGCT.py sampleDB.txt patterns.txt 10.0

    .. note:: minSup will be considered in percentage of database transactions


    **Importing this algorithm into a python program**
    ----------------------------------------------------

    .. code-block:: python

            import PAMI.frequentPattern.cuda.cuAprioriGCT as alg

            obj = alg.cuAprioriGCT(iFile, minSup)

            obj.mine()

            frequentPatterns = obj.getPatterns()

            print("Total number of Frequent Patterns:", len(frequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    **Credits:**
    -------------

                The complete program was written by Tarun Sreepada under the supervision of Professor Rage Uday Kiran.

    """

    __time = 0
    __memRSS = 0
    __memUSS = 0
    __GPU_MEM = 0
    _minSup = 0
    _finalPatterns = {}

    def __init__(self, filePath, minSup, sep):
        self._iFile = filePath
        self._sep = sep
        self._minSup = minSup
        self.__time = 0
        self.__memRSS = 0
        self.__memUSS = 0

    def __creatingItemSets(self):
        """
        Storing the complete transactions of the database/input file in a database variable
        """
        self.__Database = []
        if isinstance(self._iFile, _ab._pd.DataFrame):
            if self._iFile.empty:
                print("its empty..")
            i = self._iFile.columns.values.tolist()
            if 'Transactions' in i:
                self.__Database = self._iFile['Transactions'].tolist()

            # print(self.Database)
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                data = _ab._urlopen(self._iFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._sep)]
                    temp = [x for x in temp if x]
                    self.__Database.append(temp)
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            temp = [i.rstrip() for i in line.split(self._sep)]
                            temp = [x for x in temp if x]
                            self.__Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def __convert(self, value):
        """

        To convert the type of user specified minSup value

        :param value: user specified minSup value

        :type value: int or float or str

        :return: converted type

        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.__Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.__Database) * value)
            else:
                value = int(value)
        return value

    def compute_vertical_bitvector_data(self):
        """
        Converting database into bit vector
        """
        # ---build item to idx mapping---#
        idx = 0
        item2idx = {}
        for transaction in self.__Database:
            for item in transaction:
                if not item in item2idx:
                    item2idx[item] = idx
                    idx += 1
        idx2item = {idx: str(int(item)) for item, idx in item2idx.items()}
        # ---build vertical data---#
        vb_data = np.zeros((len(item2idx), len(self.__Database)), dtype=np.uint16)
        for trans_id, transaction in enumerate(self.__Database):
            for item in transaction:
                vb_data[item2idx[item], trans_id] = 1
        vb_data = gpuarray.to_gpu(vb_data.astype(np.uint16))
        return vb_data, idx2item

    def getRuntime(self):
        """
        Calculating the total amount of time taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """
        return self.__time

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self.__memRSS

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """
        return self.__memUSS

    def getGPUMemory(self):
        """
        To calculate the total memory consumed by GPU
        :return: return GPU memory
        :rtype: int
        """

        return self.__GPU_MEM

    def getPatterns(self):
        """
        Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._finalPatterns

    def get_numberOfPatterns(self):
        return len(self._finalPatterns)

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataFrame = {}
        data = []
        for a, b in self._finalPatterns.items():
            data.append([a, b])
            dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile):
        """
        Complete set of frequent patterns will be loaded in to an output file
        :param outFile: name of the output file
        :type outFile: csvfile
        """
        self._oFile = outFile
        writer = open(self._oFile, 'w+')
        for x, y in self._finalPatterns.items():
            if type(x) == tuple:
                pattern = ""
                for item in x:
                    pattern = pattern + str(item) + " "
                s1 = pattern + ":" + str(y)
            else:
                s1 = str(x) + ":" + str(y)
            writer.write("%s \n" % s1)

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")
    def startMine(self):
        """
        Frequent pattern mining process will start from here
        """
        self.mine()

    def mine(self):
        """
        Frequent pattern mining process will start from here
        """
        startTime = time.time()
        basePattern = {}
        final = {}

        self.__creatingItemSets()
        self._minSup = self.__convert(self._minSup)
        minSup = self._minSup
        vb_data, idx2item = self.compute_vertical_bitvector_data()

        for i in range(len(vb_data)):
            if gpuarray.sum(vb_data[i]).get() >= self._minSup:
                basePattern[idx2item[i]] = [i]
                final[idx2item[i]] = gpuarray.sum(vb_data[i]).get()

        while len(basePattern) > 0:
            temp = {}
            keysList = list(basePattern.keys())
            valuesList = list(basePattern.values())
            for i in range(len(basePattern) - 1):
                keyI = keysList[i].split(" ")
                keyI = [int(x) for x in keyI]

                for j in range(i + 1, len(basePattern)):
                    keyJ = keysList[j].split(" ")
                    keyJ = [int(x) for x in keyJ]
                    values = set(valuesList[i])
                    for val in valuesList[j]:
                        values.add(val)
                    values = list(sorted(values))
                    totalArray = vb_data[values[0]]
                    for k in range(1, len(values)):
                        totalArray = totalArray.__mul__(vb_data[values[k]])
                    support = gpuarray.sum(totalArray).get()
                    if support >= self._minSup:
                        combinedKey = " ".join(
                            str(x) for x in sorted(set(keyI) | set(keyJ)))
                        temp[combinedKey] = values
                        final[str(combinedKey)] = support
            basePattern = temp

        self.__time = time.time() - startTime
        self.__memRSS = psutil.Process(os.getpid()).memory_info().rss
        self.__memUSS = psutil.Process(os.getpid()).memory_full_info().uss
        self._finalPatterns = final
        self.__GPU_MEM = vb_data.nbytes

    def printResults(self):
        """
        This function is used to print the results
        """
        print("Total number of Coverage Patterns:", len(self.getPatterns()))
        print("GPU MEM: ", _ap.getGPUMemory())
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = cudaAprioriGCT(_ab._sys.argv[1], _ab._sys.argv[3], _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = cudaAprioriGCT(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _ap.mine()
        print("Total number of Frequent Patterns:", len(_ap.getPatterns()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("GPU MEM: ", _ap.getGPUMemory())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")




# visualizePatterns is used to visualize points produced by pattern miner .
#
#  **Importing this algorithm into a python program**
#  --------------------------------------------------------
#
#     from PAMI.extras.graph import visualizePatterns as viz
#
#     obj = viz.visualizePatterns(iFile, topk)
#
#     obj.save()
#





__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import plotly.express as px
import pandas as pd
import sys


class visualizePatterns():
    """

       :Description:   visualizePatterns is used to visualize points produced by pattern miner .

       :Attributes:

            :param file : file
                store input data as file
            :param topk : int
                Takes the value int as input

       **Importing this algorithm into a python program**
       --------------------------------------------------------
       .. code-block:: python

                from PAMI.extras.graph import visualizePatterns as viz

                obj = viz.visualizePatterns(iFile, topk)

                obj.save()

    """

    def __init__(self, file: str, topk: int) -> None:
        self.file = file
        self.topk = topk

    def visualize(self, markerSize: int = 20, zoom: int = 3, width: int = 1500, height: int = 1000) -> None:
        """
        Visualize points produced by pattern miner.

        :param markerSize: Size of the marker
        :type markerSize: int
        :param zoom: Zoom level
        :type zoom: int
        :param width: Width of the graph
        :type width: int
        :param height: Height of the graph on the screen
        :type width: int
        :return: None
        """

        long = []
        lat = []
        name = []
        color = []
        R = G = B = 0

        lines = {}
        with open(self.file, "r") as f:
            for line in f:
                lines[line] = len(line)

        lines = list(dict(sorted(lines.items(), key=lambda x: x[1])[-self.topk:]).keys())

        start = 1

        print("Number \t Pattern")
        for line in lines:

            start += 1
            if start % 3 == 0:
                R += 20
            if start % 3 == 1:
                G += 20
            if start % 3 == 2:
                B += 20
            if R > 255:
                R = 0
            if G > 255:
                G = 0
            if B > 255:
                B = 0
            RHex = hex(R)[2:]
            GHex = hex(G)[2:]
            BHex = hex(B)[2:]

            line = line.split(":")
            freq = line[-1]
            freq = "Frequency: " + freq.strip()
            line = line[:-1]
            print(str(start) + "\t" + line[0])
            points = line[0].split("\t")
            points = [x for x in points if x != ""]
            points = [x.strip("Point()") for x in points]
            for i in range(len(points)):
                temp = points[i].split()
                lat.append(float(temp[0]))
                long.append(float(temp[1]))
                name.append(freq)
                color.append("#" + RHex + GHex + BHex)
        df = pd.DataFrame({"lon": long, "lat": lat, "freq": name, "col": color})

        fig = px.scatter_mapbox(df, lat="lon", lon="lat", hover_name="freq", color="col", zoom=zoom, width=width,
                                height=height)
        fig.update_layout(mapbox_style="open-street-map")
        fig.update_traces({'marker': {'size': markerSize}})
        fig.show()


if __name__ == "__main__":
    _ap = str()
    _ap = visualizePatterns(sys.argv[1], sys.argv[2])
    _ap.visualize()



# GPFgrowth is algorithm to mine the partial periodic frequent pattern in temporal database.
#
#
# **Importing this algorithm into a python program**
# --------------------------------------------------------
#
#     from PAMI.partialPeriodicFrequentPattern.basic import GPFgrowth as alg
#
#     obj = alg.GPFgrowth(inputFile, outputFile, minSup, maxPer, minPR)
#
#     obj.startMine()
#
#     partialPeriodicFrequentPatterns = obj.getPatterns()
#
#     print("Total number of partial periodic Patterns:", len(partialPeriodicFrequentPatterns))
#
#     obj.save(oFile)
#
#     Df = obj.getPatternInDf()
#
#     memUSS = obj.getMemoryUSS()
#
#     print("Total Memory in USS:", memUSS)
#
#     memRSS = obj.getMemoryRSS()
#
#     print("Total Memory in RSS", memRSS)
#
#     run = obj.getRuntime()
#
#     print("Total ExecutionTime in seconds:", run)
#

__copyright__ = """
 Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran

"""

import deprecated
from PAMI.partialPeriodicFrequentPattern.basic.abstract import *

orderOfItem = {}

class Node:
    """
    A class used to represent the node of frequentPatternTree

    :Attributes:

        item : int
            storing item of a node
        parent : node
            To maintain the parent of every node
        child : list
            To maintain the children of node
        nodeLink : node
            To maintain the next node of node
        tidList : set
            To maintain timestamps of node

    :Methods:

        getChild(itemName)
            storing the children to their respective parent nodes
    """

    def __init__(self):
        self.item = -1
        self.parent = None
        self.child = []
        self.nodeLink = None
        self.tidList = set()

    def getChild(self, item):
        """
        :param item:
        :return: if node have node of item, then return it. if node don't have return []
        """
        for child in self.child:
            if child.item == item:
                return child
        return []


class Tree:
    """
    A class used to represent the frequentPatternGrowth tree structure

    :Attributes:

        root : node
            Represents the root node of the tree
        nodeLinks : dictionary
            storing last node of each item
        firstNodeLink : dictionary
            storing first node of each item

    :Methods:

        addTransaction(transaction,timeStamp)
            creating transaction as a branch in frequentPatternTree
        fixNodeLinks(itemName, newNode)
            add newNode link after last node of item
        deleteNode(itemName)
            delete all node of item
        createPrefixTree(path,timeStampList)
            create prefix tree by path
        createConditionalTree(PFList, minSup, maxPer, minPR, last)
            create conditional tree. Its nodes are satisfy IP / (minSup+1) >= minPR

    """
    def __init__(self):
        self.root = Node()
        self.nodeLinks = {}
        self.firstNodeLink = {}

    def addTransaction(self, transaction, tid):
        """
        add transaction into tree

        :param transaction: it represents the one transactions in database
        :type transaction: list
        :param tid: represents the timestamp of transaction
        :type tid: list
        """
        current = self.root
        for item in transaction:
            child = current.getChild(item)
            if not child:
                newNode = Node()
                newNode.item = item
                newNode.parent = current
                current.child.append(newNode)
                current = newNode
                self.fixNodeLinks(item, newNode)
            else:
                current = child
            current.tidList.add(tid)

    def fixNodeLinks(self, item, newNode):
        """
        fix node link

        :param item: it represents item name of newNode
        :type item: string
        :param newNode: it represents node which is added
        :type newNode: Node
        """
        if item in self.nodeLinks:
            lastNode = self.nodeLinks[item]
            lastNode.nodeLink = newNode
        self.nodeLinks[item] = newNode
        if item not in self.firstNodeLink:
            self.firstNodeLink[item] = newNode

    def deleteNode(self, item):
        """
        delete the node from tree

        :param item: it represents the item name of node
        :type item: str
        """
        deleteNode = self.firstNodeLink[item]
        parentNode = deleteNode.parent
        parentNode.child.remove(deleteNode)
        parentNode.child += deleteNode.child
        parentNode.tidList |= deleteNode.tidList
        for child in deleteNode.child:
            child.parent = parentNode
        while deleteNode.nodeLink:
            deleteNode = deleteNode.nodeLink
            parentNode = deleteNode.parent
            parentNode.child.remove(deleteNode)
            parentNode.child += deleteNode.child
            parentNode.tidList |= deleteNode.tidList
            for child in deleteNode.child:
                child.parent = parentNode

    def createPrefixTree(self, path, tidList):
        """
        create prefix tree by path

        :param path: it represents path to root from prefix node
        :type path: list
        :param tidList: it represents tid of each item
        :type tidList: list
        """
        currentNode = self.root
        for item in path:
            child = currentNode.getChild(item)
            if not child:
                newNode = Node()
                newNode.item = item
                newNode.parent = currentNode
                currentNode.child.append(newNode)
                currentNode = newNode
                self.fixNodeLinks(item, newNode)
            else:
                currentNode = child
            currentNode.tidList |= tidList

    def createConditionalTree(self, PFList, minSup, maxPer, minPR, last):
        """
        create conditional tree by PFlist

        :param PFList: it represents timestamp each item
        :type PFList: dict
        :param minSup: it represents minSup
        :param maxPer: it represents maxPer
        :param minPR: it represents minPR
        :param last: it represents last timestamp in database
        :return: return is PFlist which satisfy ip / (minSup+1) >= minPR
        """
        keys = list(PFList)
        for item in keys:
            ip = calculateIP(maxPer, PFList[item], last).run()
            if ip / (minSup+1) >= minPR:
                continue
            else:
                self.deleteNode(item)
                del PFList[item]
        return PFList

class calculateIP:
    """
    This class calculate ip from timestamp

    :Attributes:

    maxPer : float
        it represents user defined maxPer value
    timeStamp : list
        it represents timestamp of item
    timeStampFinal : int
        it represents last timestamp of database

    :Methods:

    run
        calculate ip from its timestamp list

    """
    def __init__(self, maxPer, timeStamp, timeStampFinal):
        self.maxPer = maxPer
        self.timeStamp = timeStamp
        self.timeStampFinal = timeStampFinal

    def run(self):
        """
        calculate ip from timeStamp list
        :return: it represents ip value
        """
        ip = 0
        self.timeStamp = sorted(list(self.timeStamp))
        if len(self.timeStamp) <= 0:
            return ip
        if self.timeStamp[0] - 0 <= self.maxPer:
            ip += 1
        for i in range(len(self.timeStamp)-1):
            if (self.timeStamp[i+1] - self.timeStamp[i]) <= self.maxPer:
                ip += 1
        if abs(self.timeStamp[(len(self.timeStamp)-1)] - self.timeStampFinal) <= self.maxPer:
            ip += 1
        return ip


class generatePFListver2:
    """
    generate time stamp list from input file

    :Attributes:

        inputFile : str
            it is input file name
        minSup : float
            user defined minimum support value
        maxPer : float
            user defined max Periodicity value
        minPR : float
            user defined min PR value
        PFList : dict
            storing timestamps each item
    """
    def __init__(self, Database, minSup, maxPer, minPR):
        self.Database = Database
        self.minSup = minSup
        self.maxPer = maxPer
        self.minPR = minPR
        self.PFList = {}
        self.tsList = {}

    def run(self):
        """
        generate PFlist
        :return: timestamps and last timestamp
        """
        global orderOfItem
        tidList = {}
        tid = 1
        for transaction in self.Database:
            timestamp = int(transaction[0])
            for item in transaction[1:]:
                if item not in self.PFList:
                    self.tsList[item] = {}
                    self.tsList[item][tid] = timestamp
                    ip = 0
                    if timestamp <= self.maxPer:
                        ip = 1
                    self.PFList[item] = [1, ip, timestamp]
                else:
                    self.tsList[item][tid] = timestamp
                    if timestamp - self.PFList[item][2] <= self.maxPer:
                        self.PFList[item][1] += 1
                    self.PFList[item][0] += 1
                    self.PFList[item][2] = timestamp
            tid += 1
            last = timestamp
        for item in self.PFList:
            if last - self.PFList[item][2] <= self.maxPer:
                self.PFList[item][1] += 1
            if self.PFList[item][1] / (self.minSup + 1) < self.minPR or self.PFList[item][0] < self.minSup:
                del self.tsList[item]
        #self.PFList = {tuple([k]): v for k, v in sorted(self.PFList.items(), key=lambda x:x[1], reverse=True)}
        tidList = {tuple([k]): v for k, v in sorted(self.tsList.items(), key=lambda x:len(x[1]), reverse=True)}
        orderOfItem = tidList.copy()
        return tidList, last

    def findSeparator(self, line):
        """
        find separator of line in database

        :param line: it represents one line in database
        :type line: list
        :return: return separator
        """
        separator = ['\t', ',', '*', '&', ' ', '%', '$', '#', '@', '!', '    ', '*', '(', ')']
        for i in separator:
            if i in line:
                return i
        return None

class generatePFTreever2:
    """
    create tree from tidList and input file

    :Attributes:

        inputFile : str
            it represents input file name
        tidList : dict
            storing tids each item
        root : Node
            it represents the root node of the tree

    :Methods:

        run
            it create tree
        find separator(line)
            find separator in the line of database

    """
    def __init__(self, Database, tidList):
        self.Database = Database
        self.tidList = tidList
        self.root = Tree()

    def run(self):
        """
        create tree from database and tidList
        :return: the root node of tree
        """
        tid = 1
        for transaction in self.Database:
            timestamp = int(transaction[0])
            transaction = [tuple([item]) for item in transaction[1:] if tuple([item]) in self.tidList]
            transaction = sorted(transaction, key=lambda x: len(self.tidList[x]), reverse=True)
            self.root.addTransaction(transaction, timestamp)
            tid += 1
        return self.root

    def findSeparator(self, line):
        """
        find separator of line in database

        :param line: it represents one line in database
        :type line: list
        :return: return separator
        """
        separator = ['\t', ',', '*', '&', ' ', '%', '$', '#', '@', '!', '    ', '*', '(', ')']
        for i in separator:
            if i in line:
                return i
        return None

class PFgrowth:
    """
    This class is pattern growth algorithm

    :Attributes:

        tree : Node
            represents the root node of prefix tree
        prefix : list
            prefix is list of prefix items
        PFList : dict
            storing time stamp each item
        minSup : float
            user defined min Support
        maxPer : float
            user defined max Periodicity
        minPR : float
            user defined min PR
        last : int
            represents last time stamp in database

    :Methods:

        run
            it is pattern growth algorithm

    """
    def __init__(self, tree, prefix, PFList, minSup, maxPer, minPR, last):
        self.tree = tree
        self.prefix = prefix
        self.PFList = PFList
        self.minSup = minSup
        self.maxPer = maxPer
        self.minPR = minPR
        self.last = last

    def run(self):
        """
        run the pattern growth algorithm
        :return: partial periodic frequent pattern in conditional pattern
        """
        global orderOfItem
        result = {}
        items = self.PFList
        if not self.prefix:
            items = reversed(self.PFList)
        for item in items:
            prefix = self.prefix.copy()
            prefix.append(item)
            PFList = {}
            prefixTree = Tree()
            prefixNode = self.tree.firstNodeLink[item]
            tidList = prefixNode.tidList
            path = []
            currentNode = prefixNode.parent
            while currentNode.item != -1:
                path.insert(0, currentNode.item)
                currentNodeItem = currentNode.item
                if currentNodeItem in PFList:
                    PFList[currentNodeItem] |= tidList
                else:
                    PFList[currentNodeItem] = tidList
                currentNode = currentNode.parent
            prefixTree.createPrefixTree(path, tidList)
            while prefixNode.nodeLink:
                prefixNode = prefixNode.nodeLink
                tidList = prefixNode.tidList
                path = []
                currentNode = prefixNode.parent
                while currentNode.item != -1:
                    path.insert(0, currentNode.item)
                    currentNodeItem = currentNode.item
                    if currentNodeItem in PFList:
                        PFList[currentNodeItem] = PFList[currentNodeItem] | tidList
                    else:
                        PFList[currentNodeItem] = tidList
                    currentNode = currentNode.parent
                prefixTree.createPrefixTree(path, tidList)
            ip = calculateIP(self.maxPer, self.PFList[item], self.last).run()
            s = len(self.PFList[item])
            if ip / (s+1) >= self.minPR and s >= self.minSup:
                result[tuple(prefix)] = [s, ip / (s+1)]
            if PFList:
                PFList = {k: v for k, v in sorted(PFList.items(), key=lambda x:len(orderOfItem[x[0]]), reverse=True)}
                PFList = prefixTree.createConditionalTree(PFList, self.minSup, self.maxPer, self.minPR, self.last)
            if PFList:
                #self.PFList = {tuple([k]): v for k, v in sorted(self.PFList.items(), key=lambda x: x[1], reverse=True)}
                obj = PFgrowth(prefixTree, prefix, PFList, self.minSup, self.maxPer, self.minPR, self.last)
                result1 = obj.run()
                result = {**result, **result1}
        return result

class GPFgrowth(partialPeriodicPatterns):
    """
    :Description:   GPFgrowth is algorithm to mine the partial periodic frequent pattern in temporal database.
    
    :Reference:   R. Uday Kiran, J.N. Venkatesh, Masashi Toyoda, Masaru Kitsuregawa, P. Krishna Reddy, Discovering partial periodic-frequent patterns in a transactional database,
                  Journal of Systems and Software, Volume 125, 2017, Pages 170-182, ISSN 0164-1212, https://doi.org/10.1016/j.jss.2016.11.035.

    :param  iFile: str :
                   Name of the Input file to mine complete set of frequent pattern's
    :param  oFile: str :
                   Name of the output file to store complete set of frequent patterns
    :param  minSup: str:
                   The user can specify minSup either in count or proportion of database size.
    :param  minPR: str:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.
    :param  maxPer: str:
                   Controls the maximum number of transactions in which any two items within a pattern can reappear.

    :param  sep: str :
                   This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.

    :Attributes:

        inputFile : file
            Name of the input file to mine complete set of frequent pattern
        minSup : float
            The user defined minSup
        maxPer : float
            The user defined maxPer
        minPR : float
            The user defined minPR
        finalPatterns : dict
            it represents to store the pattern
        runTime : float
            storing the total runtime of the mining process
        memoryUSS : float
            storing the total amount of USS memory consumed by the program
        memoryRSS : float
            storing the total amount of RSS memory consumed by the program

    :Methods:

        startMine()
            Mining process will start from here
        getPatterns()
            Complete set of patterns will be retrieved with this function
        storePatternsInFile(ouputFile)
            Complete set of frequent patterns will be loaded in to an output file
        getPatternsAsDataFrame()
            Complete set of frequent patterns will be loaded in to an output file
        getMemoryUSS()
            Total amount of USS memory consumed by the mining process will be retrieved from this function
        getMemoryRSS()
            Total amount of RSS memory consumed by the mining process will be retrieved from this function
        getRuntime()
            Total amount of runtime taken by the mining process will be retrieved from this function

    **Executing code on Terminal:**
    ----------------------------------
            Format:
                    >>> python3 GPFgrowth.py <inputFile> <outputFile> <minSup> <maxPer> <minPR>

            Examples:
                    >>> python3 GPFgrowth.py sampleDB.txt patterns.txt 10 10 0.5

    **Sample run of the importing code:**
    --------------------------------------
    ...     code-block:: python

            from PAMI.partialPeriodicFrequentPattern.basic import GPFgrowth as alg

            obj = alg.GPFgrowth(inputFile, outputFile, minSup, maxPer, minPR)

            obj.startMine()

            partialPeriodicFrequentPatterns = obj.getPatterns()

            print("Total number of partial periodic Patterns:", len(partialPeriodicFrequentPatterns))

            obj.save(oFile)

            Df = obj.getPatternInDf()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)

    **Credits:**
    --------------
            The complete program was written by Nakamura  under the supervision of Professor Rage Uday Kiran.

    """
    _partialPeriodicPatterns__iFile = ' '
    _partialPeriodicPatterns__oFile = ' '
    _partialPeriodicPatterns__sep = str()
    _partialPeriodicPatterns__startTime = float()
    _partialPeriodicPatterns__endTime = float()
    _partialPeriodicPatterns__minSup = str()
    _partialPeriodicPatterns__maxPer = str()
    _partialPeriodicPatterns__minPR = str()
    _partialPeriodicPatterns__finalPatterns = {}
    runTime = 0
    _partialPeriodicPatterns__memoryUSS = float()
    _partialPeriodicPatterns__memoryRSS = float()
    __Database = []

    def __convert(self, value):
        """
        to convert the type of user specified minSup value

        :param value: user specified minSup value
        :return: converted type
        """
        if type(value) is int:
            value = int(value)
        if type(value) is float:
            value = (len(self.__Database) * value)
        if type(value) is str:
            if '.' in value:
                value = float(value)
                value = (len(self.__Database) * value)
            else:
                value = int(value)
        return value

    def __readDatabase(self):
        self.__Database = []
        if isinstance(self.__inputFile, pd.DataFrame):
            if self.__inputFile.empty:
                print("its empty..")
            i = self.__inputFile.columns.values.tolist()
            if 'Transactions' in i:
                self.__Database = self.__inputFile['Transactions'].tolist()
            if 'Patterns' in i:
                self.__Database = self.__inputFile['Patterns'].tolist()
        if isinstance(self.__inputFile, str):
            if validators.url(self.__inputFile):
                data = urlopen(self.__inputFile)
                for line in data:
                    line.strip()
                    line = line.decode("utf-8")
                    temp = [i.rstrip() for i in line.split(self._partialPeriodicPatterns__sep)]
                    temp = [x for x in temp if x]
                    self.__Database.append(temp)
            else:
                try:
                    with open(self.__inputFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line.strip()
                            temp = [i.rstrip() for i in line.split(self._partialPeriodicPatterns__sep)]
                            temp = [x for x in temp if x]
                            self.__Database.append(temp)
                except IOError:
                    print("File Not Found")
                    quit()

    def mine(self):
        self.__inputFile = self._partialPeriodicPatterns__iFile
        self._partialPeriodicPatterns__startTime = time.time()
        self._partialPeriodicPatterns__finalPatterns = {}
        self.__readDatabase()
        self._partialPeriodicPatterns__minSup = self.__convert(self._partialPeriodicPatterns__minSup)
        self._partialPeriodicPatterns__maxPer = self.__convert(self._partialPeriodicPatterns__maxPer)
        # self.minPR = self.convert(self.minPR)
        self._partialPeriodicPatterns__finalPatterns = {}
        obj = generatePFListver2(self.__Database, self._partialPeriodicPatterns__minSup, self._partialPeriodicPatterns__maxPer, self._partialPeriodicPatterns__minPR)
        tidList, last = obj.run()
        PFTree = generatePFTreever2(self.__Database, tidList).run()
        obj2 = PFgrowth(PFTree, [], tidList, self._partialPeriodicPatterns__minSup, self._partialPeriodicPatterns__maxPer, self._partialPeriodicPatterns__minPR, last)
        self._partialPeriodicPatterns__finalPatterns = obj2.run()
        self._partialPeriodicPatterns__endTime = time.time()
        self.__runTime = self._partialPeriodicPatterns__endTime - self._partialPeriodicPatterns__startTime
        process = psutil.Process(os.getpid())
        self._partialPeriodicPatterns__memoryUSS = float()
        self._partialPeriodicPatterns__memoryRSS = float()
        self._partialPeriodicPatterns__memoryUSS = process.memory_full_info().uss
        self._partialPeriodicPatterns__memoryRSS = process.memory_info().rss


    def getMemoryUSS(self):
        """Total amount of USS memory consumed by the mining process will be retrieved from this function
        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._partialPeriodicPatterns__memoryUSS

    def getMemoryRSS(self):
        """Total amount of RSS memory consumed by the mining process will be retrieved from this function
        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._partialPeriodicPatterns__memoryRSS

    def getRuntime(self):
        """Calculating the total amount of runtime taken by the mining process
        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self.__runTime

    def getPatternsAsDataFrame(self):
        """Storing final frequent patterns in a dataframe
        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        dataframe = {}
        data = []
        for a, b in self._partialPeriodicPatterns__finalPatterns.items():
            if len(a) == 1:
                pattern = f'{a[0][0]}'
            else:
                pattern = f'{a[0][0]}'
                for item in a[1:]:
                    pattern = pattern + f' {item[0]}'
            #print(pattern)
            data.append([pattern, b[0], b[1]])
            dataframe = pd.DataFrame(data, columns=['Patterns', 'Support', 'PeriodicRatio'])
        return dataframe

    def save(self, outFile):
        """Complete set of frequent patterns will be loaded in to an output file
        :param outFile: name of the output file
        :type outFile: csv file
        """
        self.oFile = outFile
        writer = open(self.oFile, 'w+')
        for x, y in self._partialPeriodicPatterns__finalPatterns.items():
            if len(x) == 1:
                writer.write(f'{x[0][0]}:{y[0]}:{y[1]}\n')
            else:
                writer.write(f'{x[0][0]}')
                for item in x[1:]:
                    writer.write(f'\t{item[0]}')
                writer.write(f':{y[0]}:{y[1]}\n')\

    def getPatterns(self):
        """ Function to send the set of frequent patterns after completion of the mining process
        :return: returning frequent patterns
        :rtype: dict
        """
        return self._partialPeriodicPatterns__finalPatterns

    def printResults(self):
        """ this function is used to print the results
        """
        print("Total number of Partial Periodic Frequent Patterns:", len(self.getPatterns()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())

if __name__ == '__main__':
    ap = str()
    if len(sys.argv) == 6 or len(sys.argv) == 7:
        if len(sys.argv) == 7:
            ap = GPFgrowth(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5], sys.argv[6])
        if len(sys.argv) == 6:
            ap = GPFgrowth(sys.argv[1], sys.argv[3], sys.argv[4], sys.argv[5])
        ap.mine()
        print("Total number of Frequent Patterns:", len(ap.getPatterns()))
        ap.save(sys.argv[2])
        print("Total Memory in USS:", ap.getMemoryUSS())
        print("Total Memory in RSS", ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", ap.getRuntime())
    else:
        for i in [1000, 2000, 3000, 4000, 5000]:
            _ap = GPFgrowth('/Users/Likhitha/Downloads/temporal_T10I4D100K.csv', i, 500, 0.7, '\t')
            _ap.mine()
            print("Total number of Maximal Partial Periodic Patterns:", len(_ap.getPatterns()))
            _ap.save('/Users/Likhitha/Downloads/output.txt')
            print("Total Memory in USS:", _ap.getMemoryUSS())
            print("Total Memory in RSS", _ap.getMemoryRSS())
            print("Total ExecutionTime in ms:", _ap.getRuntime())
        print("Error! The number of input parameters do not match the total number of parameters provided")



# This code uses "lift" metric to extract the association rules from given frequent patterns.
#
# **Importing this algorithm into a python program**
#
#             import PAMI.AssociationRules.basic import lift as alg
#
#             obj = alg.lift(iFile, minLift)
#
#             obj.mine()
#
#             associationRules = obj.getAssociationRules()
#
#             print("Total number of Association Rules:", len(associationRules))
#
#             obj.save(oFile)
#
#             Df = obj.getPatternsAsDataFrame()
#
#             memUSS = obj.getMemoryUSS()
#
#             print("Total Memory in USS:", memUSS)
#
#             memRSS = obj.getMemoryRSS()
#
#             print("Total Memory in RSS", memRSS)
#
#             run = obj.getRuntime()
#
#             print("Total ExecutionTime in seconds:", run)
#



__copyright__ = """
Copyright (C)  2021 Rage Uday Kiran

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
     Copyright (C)  2021 Rage Uday Kiran
     
"""

from PAMI.AssociationRules.basic import abstract as _ab
from deprecated import deprecated
# increase reucursion depth
import os
import sys
sys.setrecursionlimit(10**4)
from itertools import combinations

class lift:
    """
    About this algorithm
    ====================

    :**Description**: Association Rules are derived from frequent patterns using "lift" metric.

    :**Reference**:

    :**Parameters**:    - **iFile** (*str*) -- *Name of the Input file to mine complete set of association rules*
                        - **oFile** (*str*) -- *Name of the Output file to write association rules*
                        - **minLift** (*float*) -- *Minimum lift to mine all the satisfying association rules. The user can specify the minLift in float between the range of 0 to 1.*
                        - **sep** (*str*) -- *This variable is used to distinguish items from one another in a transaction. The default seperator is tab space. However, the users can override their default separator.*

    :**Attributes**:    - **startTime** (*float*) -- *To record the start time of the mining process.*
                        - **endTime** (*float*) -- *To record the completion time of the mining process.*
                        - **finalPatterns** (*dict*) -- *Storing the complete set of patterns in a dictionary variable.*
                        - **memoryUSS** (*float*) -- *To store the total amount of USS memory consumed by the program.*
                        - **memoryRSS** (*float*) -- *To store the total amount of RSS memory consumed by the program.*


    Execution methods
    =================

    **Terminal command**

    .. code-block:: console

      Format:

      (.venv) $ python3 lift.py <inputFile> <outputFile> <minLift> <sep>

      Example Usage:

      (.venv) $ python3 lift.py sampleDB.txt patterns.txt 0.5 ' '

    .. note:: minLift can be specified in a value between 0 and 1.
    
    
    **Calling from a python program**

    .. code-block:: python

            import PAMI.AssociationRules.basic import lift as alg

            obj = alg.lift(iFile, minLift)

            obj.mine()

            associationRules = obj.getAssociationRules()

            print("Total number of Association Rules:", len(associationRules))

            obj.save(oFile)

            Df = obj.getPatternsAsDataFrame()

            memUSS = obj.getMemoryUSS()

            print("Total Memory in USS:", memUSS)

            memRSS = obj.getMemoryRSS()

            print("Total Memory in RSS", memRSS)

            run = obj.getRuntime()

            print("Total ExecutionTime in seconds:", run)


    Credits
    =======

            The complete program was written by P. Likhitha  under the supervision of Professor Rage Uday Kiran.

    """

    _minLift = float()
    _startTime = float()
    _endTime = float()
    _iFile = " "
    _oFile = " "
    _Sep = " "
    _memoryUSS = float()
    _memoryRSS = float()
    _associationRules = {}

    def __init__(self, iFile, minLift, sep):
        """
        :param iFile: input file name or path
        :type iFile: str
        :param minLift: minimum lift
        :type minLift: float
        :param sep: Delimiter of input file
        :type sep: str
        """
        self._iFile = iFile
        self._minLift = minLift
        self._associationRules = {}
        self._sep = sep

    def _readPatterns(self):
        """
        Reading the input file and storing all the frequent patterns and their support respectively in a frequentPatterns variable.
        """
        self._associationRules = {}
        if isinstance(self._iFile, _ab._pd.DataFrame):
            pattern, support = [], []
            if self._iFile.empty:
                print("its empty..")
            cols = self._iFile.columns.values.tolist()
            for col in cols:
                if 'pattern' in col.lower():
                    pattern = self._iFile[col].tolist()
                    # print("Using column: ", col, "for pattern")
                if 'support' in col.lower():
                    support = self._iFile[col].tolist()
                    # print("Using column: ", col, "for support")
            for i in range(len(pattern)):
                # if pattern[i] != tuple(): exit()
                if type(pattern[i]) != tuple:
                    raise ValueError("Pattern should be a tuple. PAMI is going through a major revision. Please raise an issue in the github repository regarding this error and provide information regarding input and algorithm.\
                                     In the meanwhile try saving the patterns to a file using (alg).save() and use the file as input. If that doesn't work, please raise an issue in the github repository.")
                s = tuple(sorted(pattern[i]))
                self._associationRules[s] = support[i]
        if isinstance(self._iFile, str):
            if _ab._validators.url(self._iFile):
                f = _ab._urlopen(self._iFile)
                for line in f:
                    line = line.strip()
                    line = line.split(':')
                    s = line[0].split(self._sep)
                    s = tuple(sorted(s))
                    
                    self._associationRules[s] = int(line[1])
            else:
                try:
                    with open(self._iFile, 'r', encoding='utf-8') as f:
                        for line in f:
                            line = line.strip()
                            line = line.split(':')
                            s = line[0].split(self._sep)
                            s = [x.strip() for x in s]
                            s = tuple(sorted(s))
                            self._associationRules[s] = int(line[1])
                except IOError:
                    print("File Not Found")
                    quit()
        # sorted(k, key=lambda x: self._frequentPatterns[x], reverse=True)
        # return k

    @deprecated("It is recommended to use 'mine()' instead of 'startMine()' for mining process. Starting from January 2025, 'startMine()' will be completely terminated.")
    def startMine(self):
        """
        Association rule mining process will start from here
        """
        self.mine()



    def mine(self):
        """
        Association rule mining process will start from here
        """
        self._startTime = _ab._time.time()
        self._readPatterns()

        keys = list(self._associationRules.keys())

        for i in range(len(self._associationRules)):
            key = self._associationRules[keys[i]]
            for idx in range(len(keys[i]) - 1, 0, -1):
                for c in combinations(keys[i], r=idx):
                    antecedent = c
                    consequent = tuple(sorted([x for x in keys[i] if x not in antecedent]))
                    # print(antecedent, consequent)
                    lift = key / (self._associationRules[antecedent]) * self._associationRules[consequent]
                    if lift >= self._minLift:
                        self._associationRules[antecedent + tuple(['->']) + keys[i]] = lift

        self._endTime = _ab._time.time()
        process = _ab._psutil.Process(_ab._os.getpid())
        self._memoryUSS = float()
        self._memoryRSS = float()
        self._memoryUSS = process.memory_full_info().uss
        self._memoryRSS = process.memory_info().rss
        print("Association rules successfully  generated from frequent patterns ")

    def getMemoryUSS(self):
        """
        Total amount of USS memory consumed by the mining process will be retrieved from this function

        :return: returning USS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryUSS

    def getMemoryRSS(self):
        """
        Total amount of RSS memory consumed by the mining process will be retrieved from this function

        :return: returning RSS memory consumed by the mining process
        :rtype: float
        """

        return self._memoryRSS

    def getRuntime(self):
        """
        Calculating the total amount of runtime taken by the mining process

        :return: returning total amount of runtime taken by the mining process
        :rtype: float
        """

        return self._endTime - self._startTime

    def getPatternsAsDataFrame(self):
        """
        Storing final frequent patterns in a dataframe

        :return: returning frequent patterns in a dataframe
        :rtype: pd.DataFrame
        """

        # dataFrame = {}
        # data = []
        # for a, b in self._finalPatterns.items():
        #     data.append([a.replace('\t', ' '), b])
        #     dataFrame = _ab._pd.DataFrame(data, columns=['Patterns', 'Support'])
        # # dataFrame = dataFrame.replace(r'\r+|\n+|\t+',' ', regex=True)
        # return dataFrame

        dataFrame = _ab._pd.DataFrame(list(self._associationRules.items()), columns=['Patterns', 'Support'])
        return dataFrame

    def save(self, outFile: str) -> None:
        """

        Complete set of frequent patterns will be loaded in to an output file

        :param outFile: name of the output file
        :type outFile: csvfile
        :return: None
        """
        with open(outFile, 'w') as f:
            for x, y in self._associationRules.items():
                x = self._sep.join(x)
                f.write(f"{x} : {y}\n")

    def getAssociationRules(self):
        """
        Function to send the set of frequent patterns after completion of the mining process

        :return: returning frequent patterns
        :rtype: dict
        """
        return self._associationRules

    def printResults(self):
        """
        Function to send the result after completion of the mining process
        """
        print("Total number of Association Rules:", len(self.getAssociationRules()))
        print("Total Memory in USS:", self.getMemoryUSS())
        print("Total Memory in RSS", self.getMemoryRSS())
        print("Total ExecutionTime in ms:", self.getRuntime())


if __name__ == "__main__":
    _ap = str()
    if len(_ab._sys.argv) == 4 or len(_ab._sys.argv) == 5:
        if len(_ab._sys.argv) == 5:
            _ap = lift(_ab._sys.argv[1], float(_ab._sys.argv[3]), _ab._sys.argv[4])
        if len(_ab._sys.argv) == 4:
            _ap = lift(_ab._sys.argv[1], _ab._sys.argv[3])
        _ap.startMine()
        _ap.mine()
        print("Total number of Association Rules:", len(_ap.getAssociationRules()))
        _ap.save(_ab._sys.argv[2])
        print("Total Memory in USS:", _ap.getMemoryUSS())
        print("Total Memory in RSS", _ap.getMemoryRSS())
        print("Total ExecutionTime in ms:", _ap.getRuntime())
    else:
        print("Error! The number of input parameters do not match the total number of parameters provided")
